#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
Neo4j-specific implementation of property graph abstraction
"""
from typing import Dict, Any, Tuple, List, Set

import logging
import shutil
import os
import tempfile
import uuid
import time
import json
import networkx as nx
from neo4j import GraphDatabase

from .abc_property_graph import ABCPropertyGraph, PropertyGraphImportException, \
    PropertyGraphQueryException, ABCGraphImporter

# to deal with intermittent APOC problems on MAC
APOC_RETRY_COUNT = 10


class Neo4jPropertyGraph(ABCPropertyGraph):
    """
    Neo4j-specific implementation of property graph abstraction
    """

    NEO4j_NONE = "None"

    def __init__(self, *, graph_id: str, importer, logger=None):
        """
        Initialize a property graph object. Logger is optional.
        :param graph_id:
        :param importer:
        :param logger:
        """
        super().__init__(graph_id=graph_id, importer=importer)
        assert isinstance(importer, Neo4jGraphImporter)
        self.driver = importer.driver
        if logger is None:
            self.log = logging.getLogger(__name__)
        else:
            self.log = logger

    def _validate_graph(self, rules_file: str) -> None:
        """ validate the graph imported in Neo4j according to a set of given Cypher rules"""
        f = open(rules_file)
        rules_dict = json.load(f)
        f.close()

        for r in rules_dict:
            with self.driver.session() as session:
                # print('Applying rule ', r['msg'])
                v = session.run(r['rule'], graphId=self.graph_id).single().value()
                # print("Rule {}, value {}".format(r['msg'], v))
                if v is False:
                    raise PropertyGraphImportException(graph_id=self.graph_id, msg=r['msg'])

    def validate_graph(self) -> None:
        """
        validate a graph from a built-in set of rules
        :return:
        """
        self.log.info(f'Applying validation rules to graph {self.graph_id}')
        self._validate_graph(os.path.join(os.path.dirname(__file__), 'data', 'graph_validation_rules.json'))
        self.log.info(f'Checking JSON properties of graph {self.graph_id}')
        self._validate_all_json_properties()

    def delete_graph(self) -> None:
        """
        Delete a graph from Neo4j
        :return:
        """
        self.log.debug(f'Deleting graph {self.graph_id}')
        with self.driver.session() as session:
            session.run('match (n:GraphNode {GraphID: $graphId })detach delete n', graphId=self.graph_id)

    def list_all_node_ids(self) -> List[str]:
        """
        List all NodeID properties of nodes in a graph
        :return:
        """
        query = "MATCH (n {GraphID: $graphId}) RETURN collect(n.NodeID) as nodeids"
        with self.driver.session() as session:
            val = session.run(query, graphId=self.graph_id).single()
            if val is None:
                raise PropertyGraphQueryException(graph_id=self.graph_id,
                                                  node_id=None, msg="Unable to list graph nodes")
            return val.data()['nodeids']

    def get_node_properties(self, *, node_id: str) -> (List[str], Dict[str, Any]):
        """
        get a tuple of labels (list) and properties (dict) of a node. Note that individual
        properties are strings (could be JSON encoded objects).
        :param node_id:
        :return: (list, dict)
        """
        assert node_id is not None
        query = "MATCH (n:GraphNode {GraphID: $graphId, NodeID: $nodeId}) RETURN labels(n), properties(n)"
        with self.driver.session() as session:
            val = session.run(query, graphId=self.graph_id, nodeId=node_id).single()
            if val is None:
                raise PropertyGraphQueryException(graph_id=self.graph_id,
                                                  node_id=node_id, msg="Unable to find node")
            labels = val.data()['labels(n)']
            labels.remove('GraphNode')
            return labels, val.data()['properties(n)']

    def get_node_json_property_as_object(self, *, node_id: str, prop_name: str) -> Any:
        """
        Return node property as a python object (applies to JSON encoded properties), will
        fail on other property values.
        :param node_id:
        :param prop_name:
        :return:
        """
        assert node_id is not None
        assert prop_name is not None
        _, node_props = self.get_node_properties(node_id=node_id)
        prop_str = node_props.get(prop_name, None)
        if prop_str is None or prop_str == self.NEO4j_NONE:
            return None
        try:
            prop_val = json.loads(prop_str)
            return prop_val
        except json.decoder.JSONDecodeError:
            raise PropertyGraphQueryException(graph_id=self.graph_id,
                                              node_id=node_id,
                                              msg=f"Unable to decode property {prop_str} as JSON.")

    def get_link_properties(self, *, node_a: str, node_b: str) -> (str, Dict[str, Any]):
        """
        get link type and properties of a link between two nodes as a tuple (no multigraphs)
        :param node_a:
        :param node_b:
        :return: (type, Dict)
        """
        assert node_a is not None
        assert node_b is not None
        query = f"MATCH (a:GraphNode {{GraphID:$graphId, NodeID:$nodeA}}) -[r]- " \
            f"(b:GraphNode {{GraphID:$graphId, NodeID:$nodeB}}) RETURN type(r), properties(r)"
        with self.driver.session() as session:
            val = session.run(query, graphId=self.graph_id, nodeA=node_a, nodeB=node_b).single()
            if val is None:
                raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_a, msg="Link doesn't exist")
            return val.data()['type(r)'], val.data()['properties(r)']

    def update_node_property(self, *, node_id: str, prop_name: str, prop_val: Any) -> None:
        """
        update a single property of a node in a graph overwriting the previous value
        :param node_id:
        :param prop_name:
        :param prop_val:
        :return:
        """
        assert node_id is not None
        assert prop_name is not None
        assert prop_val is not None
        query = f"MATCH (s:GraphNode {{GraphID: $graphId, NodeID: $nodeId}}) " \
            f"SET s+={{ {prop_name}: $propVal}} RETURN properties(s)"
        with self.driver.session() as session:
            val = session.run(query, graphId=self.graph_id, nodeId=node_id, propVal=prop_val)
            if val is None or len(val.value()) == 0:
                raise PropertyGraphQueryException(graph_id=self.graph_id,
                                                  node_id=node_id, msg="Unable to set property")

    def update_nodes_property(self, *, prop_name: str, prop_val: Any) -> None:
        """
        update a selected property on all nodes of the graph
        :param prop_name:
        :param prop_val:
        :return:
        """
        assert prop_name is not None
        assert prop_val is not None
        query = f"MATCH (s:GraphNode {{GraphID: $graphId}}) " \
                f"SET s+={{ {prop_name}: $propVal}} RETURN properties(s)"
        with self.driver.session() as session:
            val = session.run(query, graphId=self.graph_id, propVal=prop_val)
            if val is None or len(val.value()) == 0:
                raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                                  msg="Unable to set property on graph nodes")

    def update_node_properties(self, *, node_id: str, props: Dict[str, Any]) -> None:
        """
        update multiple properties on a node (value types in dictionary must be convertible to string)
        :param node_id:
        :param props:
        :return:
        """
        assert node_id is not None
        assert props is not None

        all_props = ""
        for k, v in props.items():
            all_props += f"{k}: '{v}', "
        if len(all_props) > 2:
            all_props = all_props[:-2]

        query = f"MATCH (s:GraphNode {{GraphID: $graphId, NodeID: $nodeId}}) " \
            f"SET s+= {{ {all_props} }} RETURN properties(s)"
        with self.driver.session() as session:
            val = session.run(query, graphId=self.graph_id, nodeId=node_id)
            if val is None or len(val.value()) == 0:
                raise PropertyGraphQueryException(graph_id=self.graph_id,
                                                  node_id=node_id,
                                                  msg="Unable to set properties")

    def update_link_property(self, *, node_a: str, node_b: str, kind: str, prop_name: str,
                             prop_val: Any) -> None:
        """
        update a single property of a link
        :param node_a:
        :param node_b:
        :param kind:
        :param prop_name:
        :param prop_val:
        :return:
        """
        assert node_a is not None
        assert node_b is not None
        assert kind is not None
        assert prop_name is not None
        assert prop_val is not None
        query = f"MATCH (a:GraphNode {{GraphID: $graphId, NodeID: $nodeA}}) -[r:{kind}]- " \
            f"(b:GraphNode {{GraphID: $graphId, NodeID:$nodeB}}) SET s+={{ {prop_name}: $propVal}} " \
            f"RETURN properties(r)"
        with self.driver.session() as session:
            val = session.run(query, graphId=self.graph_id, nodeA=node_a, nodeB=node_b, propVal=prop_val)
            if val is None or len(val.value()) == 0:
                raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_a,
                                                  msg="Unable to set property on a link")

    def update_link_properties(self, *, node_a: str, node_b: str, kind: str,
                               props: Dict[str, Any]) -> None:
        """
        update multiple properties of a link
        :param node_a:
        :param node_b:
        :param kind:
        :param props:
        :return:
        """
        assert node_a is not None
        assert node_b is not None
        assert kind is not None
        assert props is not None

        all_props = ""
        for k, v in props.items():
            all_props += f'{k}: "{v}", '
        if len(all_props) > 2:
            all_props = all_props[:-2]

        query = "MATCH (a:GraphNode {{GraphID: $graphId, NodeID: $nodeA}}) -[r:{kind}]- " \
            f"(b:GraphNode {{GraphID: $graphId, NodeID:$nodeB}}) SET s+= {{ {all_props} }} RETURN properties(s)"
        with self.driver.session() as session:
            val = session.run(query, graphId=self.graph_id, nodeA=node_a, nodeB=node_b)
            if val is None or len(val.value()) == 0:
                raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_a,
                                                  node_b=node_b, kind=kind,
                                                  msg="Unable to set properties on a link")

    def serialize_graph(self) -> str:
        """
        Serialize a given graph into GraphML string or return None if graph not found
        :return:
        """
        inner_query = f'match(n {{GraphID: "{self.graph_id}"}}) -[r]- (m) return n, r, m'
        # run inner query to check the graph has anything in it
        with self.driver.session() as session:
            val = session.run(inner_query)
            if val.peek() is None:
                raise PropertyGraphQueryException(graph_id=self.graph_id,
                                                  node_id=None, msg="No such graph in the database")

        query = f"with '{inner_query}' as query " \
                "CALL apoc.export.graphml.query(query, null, {stream: true, useTypes: true}) " \
                "YIELD file, source, format, nodes, relationships, properties, time, " \
                "rows, batchSize, batches, done, data " \
                "RETURN file, source, format, nodes, relationships, properties, time, " \
                "rows, batchSize, batches, done, data"

        with self.driver.session() as session:
            val = session.run(query).single()
            graph_string = val.get('data')

        if graph_string == self.NEO4j_NONE:
            return None

        # CAUTION: horrible kludge - APOC exports without indicating 'labels' is a key,
        # even though it is used, and then NetworkX refuses to import. Direct imports to
        # Neo4j work though. So we add a line to XML to declare 'labels' a key. Sigh /ib 09/12/2020
        graph_lines = graph_string.splitlines()
        graph_lines.insert(3, '<key id="labels" for="node" attr.name="labels" attr.type="string"/>\n')
        graph_string = "".join(graph_lines)
        return graph_string

    def graph_exists(self) -> bool:
        """
        Does the graph with this ID exist?
        :return:
        """
        inner_query = f'match(n {{GraphID: "{self.graph_id}"}}) -[r]- (m) return n, r, m'
        # run  query to check the graph has anything in it
        with self.driver.session() as session:
            val = session.run(inner_query)
            if val.peek() is None:
                return False
        return True

    def get_nodes_on_shortest_path(self, *, node_a: str, node_z: str, rel: str = None) -> List:
        """
        Get a list of node ids that lie on a shortest path between two nodes. Return empty
        list if no path can be found.
        :param node_a:
        :param node_z:
        :param rel:
        :return:
        """
        assert node_a is not None
        assert node_z is not None

        if rel is None:
            query = "match (a:GraphNode {GraphID: $graphId, NodeID: $nodeA}) with a match " \
                    "(z:GraphNode {GraphID: $graphId, NodeID: $nodeZ}), " \
                    "p=shortestPath((a) -[*1..]- (z)) with nodes(p) as pathnodes " \
                    "unwind pathnodes as pathnode return collect(pathnode.NodeID) as nodeids"
        else:
            query = f"match (a:GraphNode {{GraphID: $graphId, NodeID: $nodeA}}) with a match " \
                    f"(z:GraphNode {{GraphID: $graphId, NodeID: $nodeZ}}), " \
                    f"p=shortestPath((a) -[:{rel}*1..]- (z)) with nodes(p) as pathnodes " \
                    f"unwind pathnodes as pathnode return collect(pathnode.NodeID) as nodeids"

        with self.driver.session() as session:
            val = session.run(query, graphId=self.graph_id, nodeA=node_a, nodeZ=node_z).single()
            if val is None:
                return list()
            return val.data()['nodeids']

    def get_first_neighbor(self, *, node_id: str, rel: str, node_label: str) -> List[str]:
        """
        Return a list of ids of nodes of this label related via relationship. List may be empty.
        :param node_id:
        :param rel:
        :param node_label:
        :return:
        """
        assert node_id is not None
        assert rel is not None
        assert node_label is not None
        query = f"match (a:GraphNode {{GraphID: $graphId, NodeID: $nodeA}}) -[:{rel}]- " \
                f"(b:{node_label} {{ GraphID: $graphId}}) return b.NodeID"
        with self.driver.session() as session:
            val = session.run(query, graphId=self.graph_id, nodeA=node_id).value()
            if val is None:
                return list()
            return val

    def get_first_and_second_neighbor(self, *, node_id: str, rel1: str, node1_label: str,
                                      rel2: str, node2_label: str) -> List:
        """
        Return a list of 2-member lists of node ids related to this node via two specified relationships.
        List may be empty.
        :param node_id:
        :param rel1:
        :param node1_label:
        :param rel2:
        :param node2_label:
        :return:
        """
        assert node_id is not None
        assert rel1 is not None
        assert node1_label is not None
        assert rel2 is not None
        assert node2_label is not None

        query = f"match (a:GraphNode {{GraphID: $graphId, NodeID: $nodeA}}) -[:{rel1}]- "\
                f"(b:{node1_label} {{GraphID: $graphId}}) -[:{rel2}]- "\
                f"(c:{node2_label} {{GraphID: $graphId}}) return b.NodeID, c.NodeID"
        with self.driver.session() as session:
            val = session.run(query, graphId=self.graph_id, nodeA=node_id).values()
            if val is None:
                return list()
            return val

    def delete_node(self, *, node_id: str):
        """
        Delete node from a graph (incident edges automatically deleted)
        :param node_id:
        :return:
        """
        query = "match (n:GraphNode {GraphID: $graphId, NodeID: $nodeId}) " \
                "call apoc.nodes.delete(n, 10) yield value return *"
        with self.driver.session() as session:
            session.run(query, graphId=self.graph_id, nodeId=node_id).single()

    def node_exists(self, *, node_id: str, label: str):
        """
        Check if this node exists
        :param node_id:
        :param label:
        :return:
        """
        raise RuntimeError("Method node implemented")

    def add_node(self, *, node_id: str, label: str, props: Dict[str, Any]) -> None:
        raise RuntimeError("Method not implemented")

    def add_link(self, *, node_a: str, rel: str, node_b: str, props: Dict[str, Any] = None) -> None:
        raise RuntimeError("Method not implemented")

    def find_matching_nodes(self, *, other_graph) -> Set:
        """
        Return a set of node ids that match between the two graphs
        :param other_graph:
        :return:
        """
        assert other_graph is not None
        assert isinstance(other_graph, ABCPropertyGraph)
        assert other_graph.graph_exists()

        query = "match(n:GraphNode {GraphID: $graphId}) with n match " \
                "(m:GraphNode {GraphID: $graphId1}) where m.NodeID=n.NodeID " \
                "return collect(n.NodeID) as common_ids"
        with self.driver.session() as session:
            val = session.run(query, graphId=self.graph_id, graphId1=other_graph.graph_id).single()
        if val is None:
            return set()
        return set(val.data()['common_ids'])

    def merge_nodes(self, node_id: str, other_graph, merge_properties: Dict = None):
        """
        Merge two nodes of the same id belonging to two graphs. Optionally
        specify merging behavior for individual properties. Common relationships are merged.
        Information on formatting merge properties:
        https://neo4j.com/labs/apoc/4.1/graph-updates/graph-refactoring/merge-nodes/
        Example merge properties in Python:
        {
            "name":'discard', # keep property of the caller
            "age":'overwrite', # keep property of the other graph
            "kids":'combine', # make a Neo4j list of properties
            "`addr.*`": 'overwrite',
            "`.*`": 'discard'
        }
        :param node_id:
        :param other_graph:
        :param merge_properties:
        :return:
        """
        assert node_id is not None
        assert other_graph is not None
        assert other_graph.graph_exists()
        if merge_properties is not None:
            # convert properties to Neo4j format
            l = list()
            for k, v in merge_properties.items():
                l.append(str(k) + ":'" + str(v) + "'")
            merge_properties_as_string = "{" + ",".join(l) + "}"
            query = f"match (n:GraphNode {{GraphID: $graphId, NodeID: $nodeId}}), " \
                    f"(m:GraphNode {{GraphID: $graphId1, NodeID: $nodeId}}) " \
                    f"with head(collect([n, m])) as nodes " \
                    f"call apoc.refactor.mergeNodes(nodes, {{properties: {merge_properties_as_string}, " \
                    f"mergeRels: true}}) yield node return node"
        else:
            query = "match (n:GraphNode {GraphID: $graphId, NodeID: $nodeId}), " \
                    "(m:GraphNode {GraphID: $graphId1, NodeID: $nodeId}) " \
                    "with head(collect([n, m])) as nodes " \
                    "call apoc.refactor.mergeNodes(nodes, {properties: 'discard', mergeRels: true}) " \
                    "yield node return node"
        with self.driver.session() as session:
            session.run(query, graphId=self.graph_id,
                        graphId1=other_graph.graph_id,
                        nodeId=node_id).single()


class Neo4jGraphImporter(ABCGraphImporter):

    def __init__(self, *, url: str, user: str, pswd: str, import_host_dir: str, import_dir: str, logger=None):
        """
        URL of Neo4j instance, credentials and directory
        from where Neo4j can import graphs.
        :param url:
        :param user:
        :param pswd:
        :param import_host_dir: as seen outside Neo4j docker.
        :param import_dir: as seen by neo4j, set to None if neo4j not in Docker
        :param logger:
        """
        self.url = url
        self.user = user
        self.pswd = pswd
        self.import_host_dir = import_host_dir
        self.import_dir = import_dir
        try:
            self.driver = GraphDatabase.driver(self.url, auth=(user, pswd))
        except Exception as e:
            msg = f"Unable to connect to Neo4j: {str(e)}"
            raise PropertyGraphImportException(graph_id=None, msg=msg)
        if logger is None:
            self.log = logging.getLogger(__name__)
        else:
            self.log = logger

    def _prep_graph(self, graph: str, graph_id: str = None) -> Tuple[str, str, str]:
        """
        Import a graphml graph, assigning it a new unique graph ID to GraphID property
        (overwrite if there already). return the name of the file where graph is saved
        with updated GraphID and the assigned graphID
        :param graph:
        :param graph_id:
        :return:
        """
        if graph_id is None:
            graph_id = str(uuid.uuid4())

        # save to file
        with tempfile.NamedTemporaryFile(suffix="-graphml", mode='w') as f1:
            f1.write(graph)
            # read using networkx
            g = nx.read_graphml(f1.name)

        # ovewrite graph id on every node and check that NodeID is present
        for n in list(g.nodes):
            if not g.nodes[n].get('NodeID', None):
                raise PropertyGraphImportException(graph_id=graph_id,
                                                   msg="Some nodes are missing NodeID property, unable to import")
            g.nodes[n][ABCPropertyGraph.GRAPH_ID] = graph_id

        # save back to GraphML
        # where to save is determined by whether importDir is set
        dest_dir = self.import_host_dir
        if dest_dir is None:
            dest_dir = tempfile.gettempdir()

        uniq_name = str(uuid.uuid4())
        host_file_name = os.path.join(dest_dir, uniq_name)
        mapped_file_name = os.path.join(self.import_dir, uniq_name)

        nx.write_graphml(g, host_file_name)

        return graph_id, host_file_name, mapped_file_name

    def _import_graph(self, graphml_file: str, graph_id: str) -> None:
        """
        import graph into neo4j giving every node a label GraphNode
        and converting Class property into a label
        :param graphml_file:
        :param graph_id:
        :return:
        """

        assert graphml_file is not None
        assert graph_id is not None
        try:
            with self.driver.session() as session:
                self.log.debug(f"Loading graph {graph_id} into Neo4j")
                session.run(
                    'call apoc.import.graphml( $fileName, {batchSize: 10000, '
                    'readLabels: true, storeNodeIds: true } ) ',
                    fileName=graphml_file).single()
                # force one common label on all imported nodes
                self.log.debug(f"Adding GraphNode label to graph {graph_id}")
                query_string = "MATCH (n {GraphID: $graphId }) SET n:GraphNode"
                session.run(query_string, graphId=graph_id)
                # convert class property into a label as well
                self.log.debug(f"Converting Class property into Neo4j label for all nodes")
                query_string = "MATCH (n {GraphID: $graphId }) " \
                               "CALL apoc.create.addLabels([ id(n) ], [ n.Class ]) YIELD node RETURN node"
                session.run(query_string, graphId=graph_id)
        except Exception as e:
            msg = f"Neo4j APOC import error {str(e)}"
            raise PropertyGraphImportException(graph_id=graph_id, msg=msg)

    def import_graph_from_string(self, *, graph_string: str, graph_id: str = None) -> Neo4jPropertyGraph:
        """
        import graph into Neo4j from a string, assigning it a unique id
        :param graph_string:
        :param graph_id:
        :return:
        """
        assert graph_string is not None

        self.log.debug(f'Importing graph with id {graph_id}')
        try:
            assigned_id, host_file_name, mapped_file_name = self._prep_graph(graph_string, graph_id)
        except Exception as e:
            msg = f"NetworkX graph error {str(e)}"
            raise PropertyGraphImportException(graph_id=graph_id, msg=msg)

        if graph_id is not None:
            assert assigned_id == graph_id

        retry = APOC_RETRY_COUNT
        while retry > 0:
            try:
                # something in APOC prevents loading sometimes on some platforms
                self.log.debug(f"Trying to load the file {mapped_file_name}")
                self._import_graph(mapped_file_name, assigned_id)
                retry = -1
            except PropertyGraphImportException:
                self.log.warning(f"Transient error, unable to load, deleting and reimporting graph {assigned_id}")
                retry = retry - 1
                self.delete_graph(graph_id=assigned_id)
                # sleep and try again
                time.sleep(1.0)

        # remove the file
        self.log.debug(f"Unlinking temporary file {host_file_name}")
        os.unlink(host_file_name)

        if retry == 0:
            raise PropertyGraphImportException(graph_id=assigned_id,
                                               msg='Unable to load graph after multiple attempts')

        return Neo4jPropertyGraph(graph_id=assigned_id, importer=self, logger=self.log)

    def import_graph_from_string_direct(self, *, graph_string: str) -> ABCPropertyGraph:
        """
        import a graph from string without any manipulations
        :param graph_string:
        :return:
        """
        # save string to temp file
        dest_dir = self.import_host_dir
        assert dest_dir is not None
        uniq_name = str(uuid.uuid4())
        # need both host file name and what is seen from inside Docker
        host_file_name = os.path.join(dest_dir, uniq_name)
        mapped_file_name = os.path.join(self.import_dir, uniq_name)

        with open(host_file_name, 'w') as f:
            f.write(graph_string)

        # get graph id
        graph_id = self.get_graph_id(graph_file=host_file_name)

        # load file
        try:
            with self.driver.session() as session:
                session.run(
                    'call apoc.import.graphml( $fileName, {batchSize: 10000, '
                    'readLabels: true, storeNodeIds: true } ) ',
                    fileName=mapped_file_name).single()
        except Exception as e:
            msg = f"Neo4j APOC import error {str(e)}"
            raise PropertyGraphImportException(graph_id=None, msg=msg)

        # unlink temp file
        self.log.debug(f"Unlinking temporary file {host_file_name}")
        os.unlink(host_file_name)

        return Neo4jPropertyGraph(graph_id=graph_id, importer=self, logger=self.log)

    def import_graph_from_file_direct(self, *, graph_file: str) -> ABCPropertyGraph:
        """
        import a graph from file without any manipulations
        :param graph_file:
        :return:
        """
        assert graph_file is not None

        # copy file
        dest_dir = self.import_host_dir
        assert dest_dir is not None
        uniq_name = str(uuid.uuid4())
        # need both host file name and what is seen from inside Docker
        host_file_name = os.path.join(dest_dir, uniq_name)
        mapped_file_name = os.path.join(self.import_dir, uniq_name)

        shutil.copyfile(graph_file, host_file_name)

        # get graph id
        graph_id = self.get_graph_id(graph_file=host_file_name)

        # load file
        try:
            with self.driver.session() as session:
                session.run(
                    'call apoc.import.graphml( $fileName, {batchSize: 10000, '
                    'readLabels: true, storeNodeIds: true } ) ',
                    fileName=mapped_file_name).single()
        except Exception as e:
            msg = f"Neo4j APOC import error {str(e)}"
            raise PropertyGraphImportException(graph_id=None, msg=msg)

        # unlink copied file
        self.log.debug(f"Unlinking copied file {host_file_name}")
        os.unlink(host_file_name)

        return Neo4jPropertyGraph(graph_id=graph_id, importer=self, logger=self.log)

    def delete_all_graphs(self) -> None:
        """
        Delete all graphs from the database
        :return:
        """
        self.log.debug('Deleting all graphs from the database')
        with self.driver.session() as session:
            session.run('match (n) detach delete n')

    def delete_graph(self, *, graph_id: str) -> None:
        """
        Delete a graph from Neo4j
        :return:
        """
        self.log.debug(f'Deleting graph {graph_id}')
        with self.driver.session() as session:
            session.run('match (n:GraphNode {GraphID: $graphId })detach delete n', graphId=graph_id)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)

import enum

from .base_sliver import BaseSliver
from .switch_fabric import SwitchFabricInfo


@enum.unique
class ComponentType(enum.Enum):
    GPU = enum.auto()
    SmartNIC = enum.auto()
    SharedNIC = enum.auto()
    FPGA = enum.auto()
    NVME = enum.auto()

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name


class ComponentSliver(BaseSliver):

    def __init__(self):
        super().__init__()
        self.switch_fabric_info = None

    def set_switch_fabric_info(self, sf_info: SwitchFabricInfo):
        self.switch_fabric_info = sf_info

    @staticmethod
    def type_from_str(ctype: str) -> ComponentType:
        if ctype is None:
            return None
        for t in ComponentType:
            if ctype == str(t):
                return t


class AttachedComponentsInfo:
    """
    Stores attached components as a dictionary by PCI ID
    """
    def __init__(self):
        self.devices = {}

    def add_device(self, device_info: ComponentSliver):
        self.devices[device_info.pci_id] = device_info

    def remove_device(self, pci_id: str):
        if pci_id in self.devices:
            self.devices.pop(pci_id)

    def get_device(self, pci_id: str):
        return self.devices.get(pci_id, None)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)

"""
Defines various types of typed tuples: labels and capacities for things that are possible in our domain.
Labels and capacities are represented either as a single string "ltype:lval". Types are strings.
Values are simple strings for labels and integers for capacities.
Obviously ':' cannot be present in type names, however can be present in label values.
"""


import json
import importlib.resources as pkg_resources
from typing import List
from abc import ABC, abstractmethod

from . import data


class TypeValidator:
    """
    A sclass to help validate label types based on a
    JSON dictionaries
    """
    class __TypeValidator:
        def __init__(self, types_file: str):
            """
            load JSON dictionary from inside the package with possible types
            """
            types_json = pkg_resources.read_text(data, types_file)
            self.label_types = json.loads(types_json)

        def validate_type(self, atype: str) -> bool:
            """
            validate type against the dictionary
            :param atype:
            :return:
            """
            assert atype is not None
            return atype in self.label_types.keys()

        def get_type_description(self, atype: str):
            """
            get the description of a type
            :param atype:
            :return:
            """
            assert atype is not None
            return self.label_types[atype]

        def get_types(self) -> List:
            """
            returns a copy of available label types as list
            :return:
            """
            return list(self.label_types.keys())

    instances = dict()

    def __init__(self, cat: str, types_file: str):
        """
        create a type validator for category (e.g. "labels" or "capacities")
        from a given file
        :param cat:
        :param types_file:
        """
        assert cat is not None
        assert types_file is not None

        if cat not in TypeValidator.instances.keys():
            TypeValidator.instances[cat] = TypeValidator.__TypeValidator(types_file)
        else:
            pass

    def validate_type(self, cat: str, atype: str):
        """
        validate type in particular category
        :param cat:
        :param atype:
        :return:
        """
        assert cat is not None
        return self.instances[cat].validate_type(atype)

    def get_type_description(self, cat: str, atype: str):
        """
        return the description of a type in a category
        :param cat:
        :param atype:
        :return:
        """
        assert cat is not None
        assert atype is not None
        return self.instances[cat].get_type_description(atype)

    def get_types(self, cat: str):
        """
        list available types for a category
        :param cat:
        :return:
        """
        assert cat is not None
        return self.instances[cat].get_types()


class TypedTuple(ABC):
    """
    abstract typed tuple class (for labels and capacities e.g.)
    """
    LABEL_SEPARATOR = ":"

    @abstractmethod
    def __init__(self, **kwargs) -> None:
        """ assign, but validate against known types """
        # category and types file are expected to be overwritten
        # in subclasses
        self.lv = TypeValidator(self.category, self.types_file)
        if 'atype' in kwargs:
            atype = kwargs['atype']
            aval = kwargs['aval']
        elif 'fromstring' in kwargs:
            atype, aval = kwargs['fromstring'].strip().split(self.LABEL_SEPARATOR, 1)

        if self.lv.validate_type(self.category, atype):
            self.type = atype
            self.val = aval
        else:
            raise TypedTupleException(f"Tuple type {atype} is not among allowed types {self.lv.get_types()}")

    def get_type(self) -> str:
        """
        get label type
        :return:
        """
        return self.type

    def get_val(self) -> str:
        """
        get label value
        :return:
        """
        return self.val

    def get_as_string(self) -> str:
        """
        get label as string concatenation of type and value
        :return:
        """
        return ''.join(x for x in [self.type, self.LABEL_SEPARATOR, str(self.val)])

    def parse_from_string(self, alloc_str: str) -> None:
        """
        Parse allocatable from a string representatiaon
        :param alloc_str
        :return:
        """
        assert alloc_str is not None
        atype, aval = alloc_str.split(self.LABEL_SEPARATOR, 1)
        if self.lv.validate_type(self.category, atype):
            self.type = atype
            self.val = aval
        else:
            raise TypedTupleException(f"Invalid allocatable type in string {atype}, "
                                      f"allowed types {self.lv.get_types(self.category)}")

    def check_type(self, label) -> bool:
        """
        make sure we and the other label are of the same type
        :param label:
        :return:
        """
        return self.type == label.type

    def __repr__(self):
        return self.get_as_string()


class Label(TypedTuple):
    """
    Label expressed as a typed tuple
    """
    def __init__(self, **kwargs):
        self.category = "label"
        self.types_file = "label_types.json"
        super().__init__(**kwargs)


class Capacity(TypedTuple):
    """
    Capacity expressed as a typed tuple
    """
    def __init__(self, **kwargs):
        self.category = "cap"
        self.types_file = "capacity_types.json"
        super().__init__(**kwargs)


class Location(TypedTuple):
    """
    Location expressed as a typed tuple
    """

    def __init__(self, **kwargs):
        self.category = "location"
        self.types_file = "location_types.json"
        super().__init__(**kwargs)


class AllocationConstraint(TypedTuple):
    """
    Allocation constraint expressed as a typed tuple
    """
    def __init__(self, **kwargs):
        self.category = "constraint"
        self.types_file = "constraint_types.json"
        super().__init__(**kwargs)


class TypedTupleException(Exception):
    """
    Exception with a allocatable - label or capacity or other
    """
    @abstractmethod
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Typed Tuple exception: {msg}")


class CapacityException(TypedTupleException):
    """
    Exception with a capacity
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Capacity exception: {msg}")


class LabelException(TypedTupleException):
    """
    Exception with a label
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Label exception: {msg}")


class LabelOrCapacityException(TypedTupleException):
    """
    Exception with either label or capacity
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Label or Capacity exception {msg}")


class LocationException(TypedTupleException):
    """
    Exception with a location
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Label exception: {msg}")


class ConstraintException(TypedTupleException):
    """
    Exception with a constraint
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Label exception: {msg}")


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2021 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
import enum

from .base_sliver import BaseSliver
from .interface_info import InterfaceInfo


class SFLayer(enum.Enum):
    L2 = enum.auto()
    L3 = enum.auto()

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name


class SFType(enum.Enum):
    # only one subtype for now
    SwitchFabric = enum.auto()

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name


class SwitchFabricSliver(BaseSliver):
    """
    SwitchFabrics are typically invisible to everyone, but provide
    the abstraction of switching across some number of ports
    """

    def __init__(self):
        super().__init__()
        self.layer = None
        self.interface_info = None

    def set_layer(self, layer: SFLayer):
        self.layer = layer

    def get_layer(self) -> SFLayer:
        return self.layer

    @staticmethod
    def layer_from_str(layer: str) -> SFLayer:
        if layer is None:
            return None
        for t in SFLayer:
            if layer == str(t):
                return t

    @staticmethod
    def type_from_str(ntype: str) -> SFType:
        if ntype is None:
            return None
        for t in SFType:
            if ntype == str(t):
                return t


class SwitchFabricInfo:

    def __init__(self):
        self.switch_fabrics = {}

    def add_switch_fabric(self, sf_info: SwitchFabricSliver):
        self.switch_fabrics[sf_info.resource_name] = sf_info

    def remove_switch_fabric(self, name: str):
        if name in self.switch_fabrics.keys():
            self.switch_fabrics.pop(name)

    def get_switch_fabric(self, name: str):
        return self.switch_fabrics.get(name, None)



import fim.user.topology as topo
import fim.user.node as node
import fim.user.link as link
import fim.user.component as comp
import fim.user.switch_fabric as sf
import fim.slivers.network_node as nnode
import fim.slivers.network_link as nlink
import fim.slivers.attached_components as atcomp
import fim.slivers.component_catalog as cata
import fim.slivers.interface_info as iinfo
import fim.slivers.capacities_labels as caplab

# push some definitions up to simplify import management

ExperimentTopology = topo.ExperimentTopology
SubstrateTopology = topo.SubstrateTopology
Node = node.Node
Component = comp.Component
NodeType = nnode.NodeType
ComponentType = atcomp.ComponentType
ComponentCatalog = cata.ComponentCatalog
InterfaceType = iinfo.InterfaceType
Capacities = caplab.Capacities
Labels = caplab.Labels
Layer = sf.SFLayer
Link = link.Link
LinkType = nlink.LinkType
TopologyDetail = topo.TopologyDetail





#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)

from typing import Any, List
from abc import ABC, abstractmethod
import enum

from ..graph.abc_property_graph import ABCPropertyGraph


class ElementType(enum.Enum):
    EXISTING = enum.auto()
    NEW = enum.auto()


class ModelElement(ABC):
    """
    Abstract element of a model
    """
    @abstractmethod
    def __init__(self, *, name: str, node_id: str, topo: Any):
        """
        An element of topology relates to the topology object
        :param name:
        :param node_id:
        :param topo: Topology
        """
        assert name is not None
        assert node_id is not None
        assert topo is not None
        self.name = name
        self.topo = topo
        self.node_id = node_id

    def rename(self, new_name: str):
        """
        Rename the model element in the model graph
        :param new_name:
        :return:
        """
        assert new_name is not None
        self.name = new_name
        self.topo.graph_model.update_node_property(node_id=self.node_id,
                                                   prop_name=ABCPropertyGraph.PROP_NAME,
                                                   prop_val=new_name)

    def __repr__(self):
        labels, node_props = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        node_props.pop(ABCPropertyGraph.GRAPH_ID)
        if str(self.topo.__class__) == "<class 'fim.user.topology.ExperimentTopology'>":
            node_props.pop(ABCPropertyGraph.NODE_ID)
        ret = str(labels) + str(node_props)
        return ret

    def __str__(self):
        return self.__repr__()

    @abstractmethod
    def set_property(self, pname: str, pval):
        """
        Set a property of a model element
        :param pname:
        :param paval:
        :return:
        """

    @abstractmethod
    def get_property(self, pname: str) -> Any:
        """
        Get a property of a model element
        :param pname:
        :return:
        """

    @abstractmethod
    def set_properties(self, **kwargs):
        """
        Set multiple properties of a model element
        :param kwargs:
        :return:
        """


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2021 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)

from typing import Dict, Any, List
import os
import json
import uuid

from .attached_components import ComponentSliver, ComponentType
from .interface_info import InterfaceInfo, InterfaceSliver, InterfaceType
from .switch_fabric import SwitchFabricSliver, SwitchFabricInfo, SFLayer, SFType
from .capacities_labels import Capacities


class ComponentCatalog:
    """
    Reading/parsing of component catalog resource file, generation of
    component slivers based on models, search, details etc.
    """
    def __init__(self):
        pass

    @staticmethod
    def __read_catalog() -> List[Any]:
        catalog_file = os.path.join(os.path.dirname(__file__), 'data', 'component_catalog.json')
        f = open(catalog_file)
        catalog = json.load(f)
        f.close()
        assert isinstance(catalog, list)
        return catalog

    def generate_component(self, *, name: str, ctype: ComponentType, model: str,
                           switch_fabric_node_id: str = None,
                           interface_node_ids: List[str] = None) -> ComponentSliver:
        """
        Generate a component sliver with this name and model and properties, and interfaces as needed
        based on the catalog description. Interfaces if present are named
        as a concatenation of component name and port name from the catalog with
        and '_' connecting them.
        Raises CatalogException if the model is not found
        :param name: name to give to the component
        :param ctype: type of the component
        :param model:
        :param switch_fabric_node_id: if specified and if component has a switch fabric, put that there
        :param interface_node_ids: list of node ids for expected interfacaes if component has any
        :return:
        """
        assert name is not None
        assert model is not None
        catalog = self.__read_catalog()
        component_dict = None
        for c in catalog:
            if model == c['Model'] and str(ctype) == c['Type']:
                component_dict = c
                break

        if component_dict is None:
            raise CatalogException(f'Unable to find model {model} of type {ctype}in the catalog')

        cs = ComponentSliver()
        cs.set_resource_name(name)
        cs.set_resource_model(model)
        cs.set_resource_type(cs.type_from_str(component_dict['Type']))
        cs.set_details(component_dict['Details'])
        if 'Interfaces' in component_dict.keys():
            interfaces_dict = component_dict['Interfaces']
            if interface_node_ids is not None and len(interface_node_ids) != len(interfaces_dict.keys()):
                raise RuntimeError("The number of interface IDs provided is insufficient for this component "
                                   f"(need {len(interfaces_dict.keys())} instead of {len(interface_node_ids)}")
            iinfo = InterfaceInfo()
            id_index = 0
            for interface_name in interfaces_dict.keys():
                isliver = InterfaceSliver()
                isliver.set_resource_name(name + '_' + interface_name)
                isliver.set_resource_type(InterfaceType.TrunkPort)
                if interface_node_ids is not None:
                    isliver.node_id = interface_node_ids[id_index]
                else:
                    isliver.node_id = str(uuid.uuid4())
                id_index = id_index + 1
                cap = Capacities()
                # set port speed
                cap.set_fields(bw=int(interfaces_dict[interface_name]))
                isliver.set_capacities(cap=cap)
                iinfo.add_interface(isliver)
            sf = SwitchFabricSliver()
            if switch_fabric_node_id is None:
                switch_fabric_node_id = str(uuid.uuid4())
            sf.node_id = switch_fabric_node_id
            sf.set_resource_name(name + '_l2sf')
            sf.set_resource_type(SFType.SwitchFabric)
            # default to L2 for now
            sf.set_layer(SFLayer.L2)
            sf.interface_info = iinfo
            sfinfo = SwitchFabricInfo()
            sfinfo.add_switch_fabric(sf)
            cs.set_switch_fabric_info(sfinfo)
        return cs

    def component_details(self, *, model: str) -> str:
        """
        List details of a component specified by model
        :param model:
        :return:
        """
        assert model is not None
        catalog = self.__read_catalog()
        component_dict = None
        for c in catalog:
            if model == c['Model']:
                component_dict = c

        if component_dict is None:
            raise CatalogException(f'Unable to find model {model} in the catalog')
        return component_dict['Details']

    def search_catalog(self, *, ctype: ComponentType) -> Dict[str, str]:
        """
        Provide a dictionaty of matching component models and their details
        based on component type
        :param ctype:
        :return:
        """
        catalog = self.__read_catalog()
        component_dict = None
        matching_components = list()
        for c in catalog:
            if str(ctype) == c['Type']:
                matching_components.append(c)

        if len(matching_components) == 0:
            raise CatalogException(f'Unable to find type {ctype} in the catalog')
        ret = dict()
        for c in matching_components:
            ret[c['Model']] = c['Details']
        return ret


class CatalogException(Exception):
    """
    Component catalog exception
    """
    def __init__(self, msg: str):
        super().__init__(msg)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)

from typing import List, Set, Any
from abc import ABCMeta

import json
import networkx as nx
import networkx_query as nxq

from .abc_property_graph import PropertyGraphQueryException, PropertyGraphImportException, ABCPropertyGraph


class NetworkXMixin(metaclass=ABCMeta):
    """
    A mixin class offering useful functionality to NetworkX-based implementations
    """

    NETWORKX_LABEL = 'Class'

    def _find_node(self, *, node_id: str, graph_id: str = None) -> int:
        """
        Find a node matching this node ID or raise a Query Exception,
        returning an internal integer ID
        :param node_id:
        :return:
        """
        assert node_id is not None
        # this is a bit hacky - works for other graphs, not just us
        if graph_id is None:
            use_graph_id = self.graph_id
        else:
            use_graph_id = graph_id

        query_match = list(nxq.search_nodes(self.storage.get_graph(use_graph_id),
                                            {'and': [
                                                {'eq': [ABCPropertyGraph.NODE_ID, node_id]},
                                                {'eq': [ABCPropertyGraph.GRAPH_ID, use_graph_id]}
                                                ]}
                                            ))
        if len(query_match) == 0:
            raise PropertyGraphQueryException(graph_id=use_graph_id,
                                              node_id=node_id, msg="Unable to find node")
        if len(query_match) > 1:
            raise PropertyGraphQueryException(graph_id=use_graph_id,
                                              node_id=node_id, msg="Multiple matches found")
        return query_match[0]

    def _find_all_nodes(self) -> List[int]:
        """
        Find all nodes in this graph returning the list of their internal int IDs
        :return:
        """
        query_match = list(nxq.search_nodes(self.storage.get_graph(self.graph_id),
                                            {'eq': [ABCPropertyGraph.GRAPH_ID, self.graph_id]}))
        if len(query_match) == 0:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                              msg="Unable to find graph nodes")
        return query_match

    @staticmethod
    def _get_node_ids_for_list(graph: nx.Graph, nodelist: List) -> List[str] or None:
        """
        Take a list of internal int IDs and return a list of node ids for it for this graph
        :param graph:
        :param nodelist:
        :return:
        """
        assert nodelist is not None
        ret = list()
        for n in nodelist:
            # get node_ids of the path nodes
            ret.append(graph.nodes[n][ABCPropertyGraph.NODE_ID])
        return ret

    @staticmethod
    def _drop_edges_not_of_type(graph: nx.Graph, rel: str) -> None:
        """
        Take a graph and remove all edges that aren't of type/Class rel
        :param graph:
        :param rel:
        :return:
        """
        for e in graph.edges(data=True):
            # e is a tuple (source, target, property dict)
            if e[2].get(NetworkXMixin.NETWORKX_LABEL, None) != rel:
                # delete this edge
                graph.remove_edge(e[0], e[1])

    @staticmethod
    def _get_first_neighbors_via(graph: nx.Graph, real_node: int, rel: str) -> List[int]:
        """
        Get a set of integer internal IDs of nodes that are neighbors of this node
        and are connected via a specified relationship type/Class
        :param graph:
        :param node:
        :param rel:
        :return:
        """
        first_neighbors = set(graph.neighbors(real_node))
        # remove first neighbors connected via relationship type/Class that isn't rel
        neighbor_drop_list = list()
        for n in first_neighbors:
            if graph.edges[(real_node, n)].get(NetworkXMixin.NETWORKX_LABEL, None) != rel:
                neighbor_drop_list.append(n)

        return list(first_neighbors.difference(neighbor_drop_list))

    @staticmethod
    def _filter_nodes_by_label(graph: nx.Graph, nodeset: Set[int], node_label: str) -> List[int]:
        """
        Return a new list with only members that have the matching label
        :param graph:
        :param nodeset:
        :param node_label:
        :return:
        """
        droplist = list()
        for n in nodeset:
            if graph.nodes[n].get(NetworkXMixin.NETWORKX_LABEL, None) != node_label:
                droplist.append(n)
        return list(set(nodeset).difference(droplist))

    @staticmethod
    def _collect_nodeids(graph: nx.Graph) -> Set[str]:
        """
        Collect node ids from a NetworkX graph object as a set
        :param graph:
        :return:
        """
        assert isinstance(graph, nx.Graph)
        nodeids = set()
        for n in list(graph.nodes):
            nodeids.add(graph.nodes[n][ABCPropertyGraph.NODE_ID])
        return nodeids


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
Abstract definition of ASM (Abstract Slice Model) functionality
"""

from typing import List, Any, Dict
from abc import ABCMeta, abstractmethod

from fim.slivers.network_node import NodeSliver
from fim.slivers.attached_components import ComponentSliver
from fim.slivers.switch_fabric import SwitchFabricSliver
from fim.slivers.interface_info import InterfaceSliver
from fim.slivers.capacities_labels import Capacities, Labels
from fim.slivers.network_link import NetworkLinkSliver
from fim.graph.abc_property_graph import ABCPropertyGraph


class ABCASMMixin(metaclass=ABCMeta):
    """
    Interface for an ASM Mixin on top of a property graph
    """
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'get_all_network_nodes') and
                callable(subclass.get_all_network_nodes) or NotImplemented)

    @abstractmethod
    def check_node_unique(self, *, label: str, name: str):
        """
        Check no other node of this class/label and name exists
        :param label:
        :param name:
        :return:
        """

    @abstractmethod
    def get_all_network_nodes(self) -> List[str]:
        """
        Get a list of nodes IDs in a slice model
        :return:
        """

    @abstractmethod
    def get_all_network_links(self) -> List[str]:
        """
        Get a list of link node ids in a slice model
        :return:
        """

    @abstractmethod
    def get_all_network_node_components(self, parent_node_id: str) -> List[str]:
        """
        Get a list of node ids of components that are connected to a parent node
        :param parent_node_id:
        :return:
        """

    @abstractmethod
    def get_all_node_or_component_connection_points(self, parent_node_id: str) -> List[str]:
        """
        Get a list of interfaces attached to switch fabrics
        :param parent_node_id:
        :return:
        """

    @abstractmethod
    def get_all_network_node_or_component_sfs(self, parent_node_id: str) -> List[str]:
        """
        Get a list of node or component switch fabrics
        :param parent_node_id:
        :return:
        """

    @abstractmethod
    def get_all_sf_connection_points(self, parent_node_id: str) -> List[str]:
        """
        Get a list of interfaces as children of a switch fabric
        :param parent_node_id:
        :return:
        """

    @abstractmethod
    def get_all_link_interfaces(self, link_id: str) -> List[str]:
        """
        Get a list of interface node ids attached to this link
        :param link_id:
        :return:
        """

    @abstractmethod
    def find_node_by_name(self, node_name: str, label: str) -> str:
        """
        Get node id of node based on its name and class/label. Throw
        exception if multiple matches found.
        :param node_name:
        :param label: node label or class of the node
        :return:
        """

    @abstractmethod
    def remove_network_node_with_components_sfs_cps_and_links(self, node_id: str):
        """
        Remove a network node and all of its components from the graph
        :param node_id:
        :return:
        """

    @abstractmethod
    def remove_component_with_sfs_cps_and_links(self, node_id: str):
        """
        Remove a component with subtending switch fabrics, connection points and links
        :param node_id:
        :return:
        """

    @abstractmethod
    def remove_sf_with_cps_and_links(self, node_id: str):
        """
        Remove a SwitchFabric with subtending connection points and links
        :param node_id:
        :return:
        """

    @abstractmethod
    def remove_cp_and_links(self, node_id: str):
        """
        Remove connection point with subtending parent connection points and links
        :param node_id:
        :return:
        """

    @abstractmethod
    def add_network_node_sliver(self, *, sliver: NodeSliver):
        """
        Use node sliver to add node, components and interfaces to the graph
        :param sliver:
        :return:
        """

    @abstractmethod
    def add_network_link_sliver(self, *, lsliver: NetworkLinkSliver, interfaces: List[str]):
        """
        Add a link sliver node connecting the specified interface nodes (listed by node_id)
        :param lsliver:
        :param interfaces:
        :return:
        """

    @abstractmethod
    def add_component_sliver(self, *, parent_node_id: str, component: ComponentSliver):
        """
        Add Component (PCI device) to a node with node_id
        :param parent_node_id:
        :param component:
        :return:
        """

    @abstractmethod
    def add_switch_fabric_sliver(self, *, parent_node_id: str, switch_fabric: SwitchFabricSliver):
        """
        Add switch fabric to component or node
        :param parent_node_id:
        :param switch_fabric:
        :return:
        """

    @abstractmethod
    def add_interface_sliver(self, *, parent_node_id: str, interface: InterfaceSliver):
        """
        Add switching fabric (if needed) and an interface of the component
        :param parent_node_id:
        :param interface:
        :return:
        """

    @abstractmethod
    def find_component_by_name(self, *, parent_node_id: str, component_name: str) -> str:
        """
        Find a component by name as a child of a node
        :param parent_node_id:
        :param component_name:
        :return:
        """

    @abstractmethod
    def find_sf_by_name(self, *, parent_node_id: str, sfname: str) -> str:
        """
        Find a switch fabric by name as a child of a node or component
        :param parent_node_id:
        :param sfname:
        :return:
        """

    @abstractmethod
    def find_connection_point_by_name(self, *, parent_node_id: str, iname: str) -> str:
        """
        Find an interface by name as a child of switch fabric
        :param parent_node_id:
        :param iname:
        :return:
        """

    @staticmethod
    def node_sliver_from_graph_properties_dict(d: Dict[str, str]) -> NodeSliver:
        n = NodeSliver()
        if d.get(ABCPropertyGraph.PROP_IMAGE_REF, None) is None:
            image_ref = None
            image_type = None
        else:
            image_ref, image_type = d[ABCPropertyGraph.PROP_IMAGE_REF].split(',')
        n.set_properties(resource_name=d.get(ABCPropertyGraph.PROP_NAME, None),
                         resource_type=n.type_from_str(d.get(ABCPropertyGraph.PROP_TYPE, None)),
                         capacities=Capacities().from_json(d.get(ABCPropertyGraph.PROP_CAPACITIES, None)),
                         labels=Labels().from_json(d.get(ABCPropertyGraph.PROP_LABELS, None)),
                         site=d.get(ABCPropertyGraph.PROP_SITE, None),
                         image_ref=image_ref,
                         image_type=image_type,
                         management_ip=d.get(ABCPropertyGraph.PROP_MGMT_IP, None),
                         allocation_constraints=d.get(ABCPropertyGraph.PROP_ALLOCATION_CONSTRAINTS, None),
                         service_endpoint=d.get(ABCPropertyGraph.PROP_SERVICE_ENDPOINT, None),
                         details=d.get(ABCPropertyGraph.PROP_DETAILS, None)
                         )
        return n

    @staticmethod
    def link_sliver_from_graph_properties_dict(d: Dict[str, str]) -> NetworkLinkSliver:
        n = NetworkLinkSliver()
        n.set_properties(resource_name=d.get(ABCPropertyGraph.PROP_NAME, None),
                         resource_type=n.type_from_str(d.get(ABCPropertyGraph.PROP_TYPE, None)),
                         capacities=Capacities().from_json(d.get(ABCPropertyGraph.PROP_CAPACITIES, None)),
                         labels=Labels().from_json(d.get(ABCPropertyGraph.PROP_LABELS, None)),
                         layer=n.layer_from_str(d.get(ABCPropertyGraph.PROP_LAYER, None)),
                         technology=d.get(ABCPropertyGraph.PROP_TECHNOLOGY, None),
                         details=d.get(ABCPropertyGraph.PROP_DETAILS, None)
                         )
        return n

    @staticmethod
    def component_sliver_from_graph_properties_dict(d: Dict[str, str]) -> ComponentSliver:
        """
        Create component sliver from node graph properties
        :param d:
        :return:
        """
        cs = ComponentSliver()
        cs.set_properties(name=d.get(ABCPropertyGraph.PROP_NAME, None),
                          resource_type=cs.type_from_str(d.get(ABCPropertyGraph.PROP_TYPE, None)),
                          capacities=Capacities().from_json(d.get(ABCPropertyGraph.PROP_CAPACITIES, None)),
                          labels=Labels().from_json(d.get(ABCPropertyGraph.PROP_LABELS, None)),
                          details=d.get(ABCPropertyGraph.PROP_DETAILS, None)
                          )
        return cs

    @staticmethod
    def switch_fabric_sliver_from_graph_properties_dict(d: Dict[str, str]) -> SwitchFabricSliver:
        """
        SwitchFabric sliver from node graph properties
        :param d:
        :return:
        """
        sf = SwitchFabricSliver()
        sf.set_properties(name=d.get(ABCPropertyGraph.PROP_NAME, None),
                          resource_type=sf.type_from_str(d.get(ABCPropertyGraph.PROP_TYPE, None)),
                          capacities=Capacities().from_json(d.get(ABCPropertyGraph.PROP_CAPACITIES, None)),
                          labels=Labels().from_json(d.get(ABCPropertyGraph.PROP_LABELS, None)),
                          layer=sf.layer_from_str(d.get(ABCPropertyGraph.PROP_LAYER, None)),
                          details=d.get(ABCPropertyGraph.PROP_DETAILS, None)
                          )
        return sf

    @staticmethod
    def interface_sliver_from_graph_properties_dict(d: Dict[str, str]) -> InterfaceSliver:
        """
        Interface sliver from node graph properties
        :param d:
        :return:
        """
        isl = InterfaceSliver()
        isl.set_properties(name=d.get(ABCPropertyGraph.PROP_NAME, None),
                           resource_type=isl.type_from_str(d.get(ABCPropertyGraph.PROP_TYPE, None)),
                           capacities=Capacities().from_json(d.get(ABCPropertyGraph.PROP_CAPACITIES, None)),
                           labels=Labels().from_json(d.get(ABCPropertyGraph.PROP_LABELS, None)),
                           details=d.get(ABCPropertyGraph.PROP_DETAILS, None)
                           )
        return isl

    @staticmethod
    def node_sliver_to_graph_properties_dict(sliver: NodeSliver) -> Dict[str, str]:
        """
        This method knows how to map sliver fields to graph properties
        :param self:
        :param sliver:
        :return:
        """
        prop_dict = dict()

        if sliver.resource_name is not None:
            prop_dict[ABCPropertyGraph.PROP_NAME] = sliver.resource_name
        if sliver.resource_type is not None:
            prop_dict[ABCPropertyGraph.PROP_TYPE] = str(sliver.resource_type)
        if sliver.capacities is not None:
            prop_dict[ABCPropertyGraph.PROP_CAPACITIES] = sliver.capacities.to_json()
        if sliver.labels is not None:
            prop_dict[ABCPropertyGraph.PROP_LABELS] = sliver.labels.to_json()
        if sliver.site is not None:
            prop_dict[ABCPropertyGraph.PROP_SITE] = sliver.site
        if sliver.image_ref is not None and sliver.image_type is not None:
            prop_dict[ABCPropertyGraph.PROP_IMAGE_REF] = sliver.image_ref + ',' + str(sliver.image_type)
        if sliver.management_ip is not None:
            prop_dict[ABCPropertyGraph.PROP_MGMT_IP] = str(sliver.management_ip)
        if sliver.allocation_constraints is not None:
            prop_dict[ABCPropertyGraph.PROP_ALLOCATION_CONSTRAINTS] = sliver.allocation_constraints
        if sliver.service_endpoint is not None:
            prop_dict[ABCPropertyGraph.PROP_SERVICE_ENDPOINT] = str(sliver.service_endpoint)
        if sliver.details is not None:
            prop_dict[ABCPropertyGraph.PROP_DETAILS] = sliver.details
        return prop_dict

    @staticmethod
    def link_sliver_to_graph_properties_dict(sliver: NetworkLinkSliver) -> Dict[str, str]:
        """
        This method knows how to map sliver fields to graph properties
        :param sliver:
        :return:
        """
        prop_dict = dict()

        if sliver.resource_name is not None:
            prop_dict[ABCPropertyGraph.PROP_NAME] = sliver.resource_name
        if sliver.resource_type is not None:
            prop_dict[ABCPropertyGraph.PROP_TYPE] = str(sliver.resource_type)
        if sliver.capacities is not None:
            prop_dict[ABCPropertyGraph.PROP_CAPACITIES] = sliver.capacities.to_json()
        if sliver.labels is not None:
            prop_dict[ABCPropertyGraph.PROP_LABELS] = sliver.labels.to_json()
        if sliver.layer is not None:
            prop_dict[ABCPropertyGraph.PROP_LAYER] = str(sliver.layer)
        if sliver.technology is not None:
            prop_dict[ABCPropertyGraph.PROP_TECHNOLOGY] = str(sliver.technology)
        if sliver.details is not None:
            prop_dict[ABCPropertyGraph.PROP_DETAILS] = sliver.details

        return prop_dict

    @staticmethod
    def component_sliver_to_graph_properties_dict(sliver: ComponentSliver) -> Dict[str, str]:
        """
        This method knows how to map component sliver fields to graph properties
        :param self:
        :param sliver:
        :return:
        """
        prop_dict = dict()

        if sliver.resource_name is not None:
            prop_dict[ABCPropertyGraph.PROP_NAME] = sliver.resource_name
        if sliver.resource_type is not None:
            prop_dict[ABCPropertyGraph.PROP_TYPE] = str(sliver.resource_type)
        if sliver.capacities is not None:
            prop_dict[ABCPropertyGraph.PROP_CAPACITIES] = sliver.capacities.to_json()
        if sliver.labels is not None:
            prop_dict[ABCPropertyGraph.PROP_LABELS] = sliver.labels.to_json()
        if sliver.details is not None:
            prop_dict[ABCPropertyGraph.PROP_DETAILS] = sliver.details

        return prop_dict

    @staticmethod
    def switch_fabric_sliver_to_graph_properties_dict(sliver: SwitchFabricSliver) -> Dict[str, str]:
        """
        This method knows how to map switch fabric sliver fields to graph properties
        :param self:
        :param sliver:
        :return:
        """
        prop_dict = dict()

        if sliver.resource_name is not None:
            prop_dict[ABCPropertyGraph.PROP_NAME] = sliver.resource_name
        if sliver.resource_type is not None:
            prop_dict[ABCPropertyGraph.PROP_TYPE] = str(sliver.resource_type)
        if sliver.capacities is not None:
            prop_dict[ABCPropertyGraph.PROP_CAPACITIES] = sliver.capacities.to_json()
        if sliver.labels is not None:
            prop_dict[ABCPropertyGraph.PROP_LABELS] = sliver.labels.to_json()
        if sliver.layer is not None:
            prop_dict[ABCPropertyGraph.PROP_LAYER] = str(sliver.layer)
        if sliver.details is not None:
            prop_dict[ABCPropertyGraph.PROP_DETAILS] = sliver.details

        return prop_dict

    @staticmethod
    def interface_sliver_to_graph_properties_dict(sliver: InterfaceSliver) -> Dict[str, str]:
        """
        This method knows how to map interface sliver fields to graph properties
        :param self:
        :param sliver:
        :return:
        """
        prop_dict = dict()

        if sliver.resource_name is not None:
            prop_dict[ABCPropertyGraph.PROP_NAME] = sliver.resource_name
        if sliver.resource_type is not None:
            prop_dict[ABCPropertyGraph.PROP_TYPE] = str(sliver.resource_type)
        if sliver.capacities is not None:
            prop_dict[ABCPropertyGraph.PROP_CAPACITIES] = sliver.capacities.to_json()
        if sliver.labels is not None:
            prop_dict[ABCPropertyGraph.PROP_LABELS] = sliver.labels.to_json()
        if sliver.details is not None:
            prop_dict[ABCPropertyGraph.PROP_DETAILS] = sliver.details

        return prop_dict



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)

from typing import Any, List

import uuid

from .model_element import ModelElement, ElementType

from ..slivers.interface_info import InterfaceType, InterfaceSliver
from ..graph.abc_property_graph import ABCPropertyGraph


class Interface(ModelElement):

    def __init__(self, *, name: str, node_id: str = None, topo: Any,
                 etype: ElementType = ElementType.EXISTING,
                 parent_node_id: str = None,
                 itype: InterfaceType = None, **kwargs):
        """
        Don't call this method yourself, call node.add_interface()
        node_id will be generated if not provided for experiment topologies

        :param name:
        :param node_id:
        :param topo:
        :param etype: is this supposed to exist or new should be created
        :param parent_node_id: parent switch fabric when creating new interface
        :param itype: node type if it is new
        :param kwargs: any additional properties
        """
        assert name is not None
        assert topo is not None

        if etype == ElementType.NEW:
            # cant use isinstance as it would create circular import dependencies
            # node id myst be specified for new nodes in substrate topologies
            if str(topo.__class__) == "<class 'fim.user.topology.SubstrateTopology'>" and \
                    node_id is None:
                raise RuntimeError("When adding new nodes to substrate topology nodes you must specify static Node ID")
            if node_id is None:
                node_id = str(uuid.uuid4())
            super().__init__(name=name, node_id=node_id, topo=topo)
            if itype is None:
                raise RuntimeError("When creating interfaces you must specify InterfaceType")

            self.node_id = node_id
            sliver = InterfaceSliver()
            sliver.node_id = self.node_id
            sliver.set_resource_name(self.name)
            sliver.set_resource_type(itype)
            sliver.set_properties(**kwargs)

            self.topo.graph_model.add_interface_sliver(parent_node_id=parent_node_id, interface=sliver)
        else:
            super().__init__(name=name, node_id=node_id, topo=topo)
            # check that this node exists
            existing_node_id = self.topo.\
                graph_model.find_node_by_name(node_name=name,
                                              label=str(ABCPropertyGraph.CLASS_ConnectionPoint))
            if node_id is not None and existing_node_id != node_id:
                raise RuntimeError("Existing node id does not match provided. "
                                   "In general you shouldn't need to specify node id for existing interfaces.")
            self.node_id = node_id

    def add_child_interface(self):
        raise RuntimeError("Not implemented")

    def get_property(self, pname: str) -> Any:
        """
        Retrieve a interface property
        :param pname:
        :return:
        """
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        if_sliver = self.topo.graph_model.interface_sliver_from_graph_properties_dict(node_properties)
        return if_sliver.get_property(pname)

    def set_property(self, pname: str, pval: Any):
        """
        Set a interface property
        :param pname:
        :param pval:
        :return:
        """
        if_sliver = InterfaceSliver()
        if_sliver.set_property(prop_name=pname, prop_val=pval)
        # write into the graph
        prop_dict = self.topo.graph_model.interface_sliver_to_graph_properties_dict(if_sliver)
        self.topo.graph_model.update_node_properties(node_id=self.node_id, props=prop_dict)

    def set_properties(self, **kwargs):
        """
        Set multiple properties of the interface
        :param kwargs:
        :return:
        """
        if_sliver = InterfaceSliver()
        if_sliver.set_properties(**kwargs)
        # write into the graph
        prop_dict = self.topo.graph_model.interface_sliver_to_graph_properties_dict(if_sliver)
        self.topo.graph_model.update_node_properties(node_id=self.node_id, props=prop_dict)

    @staticmethod
    def list_properties() -> List[str]:
        return InterfaceSliver.list_properties()

    def __repr__(self):
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        if_sliver = self.topo.graph_model.interface_sliver_from_graph_properties_dict(node_properties)
        return if_sliver.__repr__()

    def __str__(self):
        return self.__repr__()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
Abstract definition of ADM (Aggregate Delegation Model) functionality
"""

from abc import ABCMeta, abstractmethod


class ABCADMMixin(metaclass=ABCMeta):
    """
    Interface for an ADM Mixin on top of a property graph
    """
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'rewrite_delegations') and
                callable(subclass.rewrite_delegations) or NotImplemented)

    @abstractmethod
    def rewrite_delegations(self, *, real_adm_id: str = None) -> None:
        """
        Rewrite label and capacity delegations on all nodes to be dictionaries
        referenced by ADM graph id. Note that external code should
        not interact with delegations on ADM graphs.
        Sometimes ADMs are cloned into temporary graphs so the method provides
        a way to pass original ADM id as an option.
        :param real_adm_id:
        :return:
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import enum
import ipaddress

from .base_sliver import BaseSliver
from .attached_components import AttachedComponentsInfo
from .interface_info import InterfaceInfo
from .switch_fabric import SwitchFabricInfo


class NodeType(enum.Enum):
    Server = enum.auto()
    VM = enum.auto()
    Container = enum.auto()
    Switch = enum.auto()

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name


class NodeSliver(BaseSliver):

    def __init__(self):
        super().__init__()
        self.management_ip = None
        self.attached_components_info = None
        self.allocation_constraints = None
        self.image_type = None
        self.image_ref = None
        self.service_endpoint = None
        self.switch_fabric_info = None
        self.site = None

    #
    # Setters are only needed for things we want users to be able to set
    #
    def set_management_ip(self, management_ip: str):
        if management_ip is None:
            self.management_ip = None
        else:
            self.management_ip = ipaddress.ip_address(management_ip)

    def get_management_ip(self) -> str:
        return self.management_ip

    def set_allocation_constraints(self, allocation_constraints: str):
        self.allocation_constraints = allocation_constraints

    def get_allocation_constraints(self) -> str:
        return self.allocation_constraints

    def set_image_type(self, image_type: str):
        self.image_type = image_type

    def get_image_type(self) -> str:
        return self.image_type

    def set_image_ref(self, image_ref: str):
        self.image_ref = image_ref

    def get_image_ref(self) -> str:
        return self.image_ref

    def set_service_endpoint(self, service_endpoint: str):
        self.service_endpoint = service_endpoint

    def get_service_endpoint(self) -> str:
        return self.service_endpoint

    def set_site(self, site: str):
        self.site = site

    def get_site(self) -> str:
        return self.site

    @staticmethod
    def type_from_str(ntype: str) -> NodeType or None:
        if ntype is None:
            return None
        for t in NodeType:
            if ntype == str(t):
                return t



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
Neo4j implementation of ADM (Aggregate Delegation Model) functionality.
"""

import uuid
import json

from ..neo4j_property_graph import Neo4jPropertyGraph, Neo4jGraphImporter
from .abc_adm import ABCADMMixin


class Neo4jADMGraph(Neo4jPropertyGraph, ABCADMMixin):

    def __init__(self, *, graph_id: str = None, importer: Neo4jGraphImporter, logger=None):
        """
        Initialize CBM either from existing ID or generate new graph id
        :param graph_id:
        :param importer:
        """
        if graph_id is None:
            graph_id = str(uuid.uuid4())

        super().__init__(graph_id=graph_id, importer=importer, logger=logger)

    def rewrite_delegations(self, *, real_adm_id: str = None) -> None:
        """
        Rewrite label and capacity delegations on all nodes to be dictionaries
        referenced by ADM graph id. Note that external code should
        not interact with delegations on ADM graphs.
        Sometimes ADMs are cloned into temporary graphs so the method provides
        a way to pass original ADM id as an option.
        :param real_adm_id:
        :return:
        """
        for node_id in self.list_all_node_ids():
            _, node_props = self.get_node_properties(node_id=node_id)
            props_modified = False
            for delegation_prop_name in [self.PROP_LABEL_DELEGATIONS, self.PROP_CAPACITY_DELEGATIONS]:
                if delegation_prop_name in node_props.keys():
                    if node_props[delegation_prop_name] == self.NEO4j_NONE:
                        continue
                    props_modified = True
                    # turn delegation into Python object
                    delegation = json.loads(node_props[delegation_prop_name])
                    new_delegation = dict()
                    if real_adm_id is None:
                        new_delegation[self.graph_id] = delegation
                    else:
                        new_delegation[real_adm_id] = delegation
                    new_delegation_json = json.dumps(new_delegation)
                    node_props[delegation_prop_name] = new_delegation_json
            if props_modified:
                # write back
                self.update_node_properties(node_id=node_id, props=node_props)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
Label and capacity delegations and pools. Each delegation knows which node it is
defined on, each pool knows which node defined it and which nodes it
applies to and can be set/queried for that information.
"""

from enum import Enum
from typing import List, Dict, Set
import json

from fim.graph.abc_property_graph import ABCPropertyGraph


class DelegationType(Enum):
    CAPACITY = 1
    LABEL = 2


class Delegation:
    """Label and capacity delegation. Only on individual resources.
    Pools are used to define delegations across multiple resources.
    """
    def __init__(self, atype: DelegationType, defined_on: str, delegation_id: str):
        """
        Define a delegation of resources on a given node
        :param atype:
        :param defined_on:
        :param delegation_id:
        """
        assert atype is not None
        assert defined_on is not None
        assert delegation_id is not None
        self.type = atype
        self.on_ = defined_on
        self.delegation_id = delegation_id
        self.delegation_details = None

    def set_details(self, *, dele_dict: Dict) -> None:
        """
        set details of the delegation dictionary, removing the delegation
        identifier field from dictionary
        :param dele_dict:
        :return:
        """
        assert dele_dict is not None
        # pop the delegation field
        dele_dict.pop(ABCPropertyGraph.FIELD_DELEGATION, None)
        self.delegation_details = dele_dict

    def get_details(self) -> Dict:
        """
        get delegation details as a dictionary (minus the delegation id)
        :return:
        """
        return self.delegation_details

    def __repr__(self) -> str:
        return f"{self.type} delegation delegated to {self.delegation_id}: " \
               f"{self.on_} with {self.delegation_details} "


class Delegations:
    """
    Multiple delegations references by delegation id
    """
    def __init__(self, atype: DelegationType):
        self.type = atype
        self.delegations = {}

    def add_delegation(self, *, delegation: Delegation) -> None:
        """
        Append a well-formed delegation to the list of delegations matching
        this id
        :param delegation:
        :return:
        """
        assert delegation.delegation_id is not None
        if self.delegations.get(delegation.delegation_id, None) is None:
            self.delegations[delegation.delegation_id] = []

        self.delegations[delegation.delegation_id].append(delegation)

    def get_by_delegation_id(self, *, delegation_id: str) -> List[Delegation]:
        """
        retrieve a list of delegations by their id or None
        :param delegation_id:
        :return:
        """
        assert delegation_id is not None
        return self.delegations.get(delegation_id, None)

    def get_delegation_ids(self) -> Set:
        """
        get a set of all delegation ids
        :return:
        """
        return set(self.delegations.keys())

    def get_node_ids(self, *, delegation_id: str) -> Set:
        """
        return a set of nodes ids for a given delegation id
        :param delegation_id:
        :return:
        """
        assert delegation_id is not None
        ret = set()
        if delegation_id not in self.delegations.keys():
            return ret
        for delegation in self.delegations[delegation_id]:
            ret.add(delegation.on_)
        return ret

    def __repr__(self):
        return f"{self.delegations}"


class Pool:
    """
    Label and capacity pools. Each pool knows which node defined it and which nodes it
    applies to and can be set/queried for that information.
    """
    def __init__(self, *, atype: DelegationType, pool_id: str, delegation_id: str = None,
                 defined_on: str = None, defined_for: List[str] = None):
        """
        Define a label or capacity pool and optionally specify where it is defined and for which nodes
        and what resources fall into it
        :param atype:
        :param pool_id:
        :param defined_on:
        :param defined_for:
        :param delegation_id:
        """
        assert atype is not None
        assert pool_id is not None
        self.type = atype
        self.on_ = defined_on
        self.delegation_id = delegation_id
        if defined_for is None:
            self.for_ = []
        else:
            self.for_ = defined_for
        self.pool_id = pool_id
        self.pool_details = None

    def set_defined_on(self, *, node_id: str) -> None:
        """
        Set ID of node on which this pool is defined
        :param node_id:
        :return:
        """
        assert node_id is not None
        self.on_ = node_id

    def set_defined_for(self, *, node_id_list: List[str]) -> None:
        """
        set/replace the pool's defined_for list of nodes
        :param node_id_list:
        :return:
        """
        assert node_id_list is not None
        assert len(node_id_list) != 0

        self.for_ = node_id_list

    def add_defined_for(self, *, node_ids) -> None:
        """
        Add to the list of nodes for which the pool is defined for.
        Individual items or lists
        :param node_ids:
        :return:
        """
        assert node_ids is not None
        if isinstance(node_ids, str):
            self.for_.append(node_ids)
        elif isinstance(node_ids, list):
            self.for_.extend(node_ids)

    def set_pool_details(self, *, pool_dict: Dict) ->None:
        """
        set details of the pool dictionary, removing delegation
        identifier field from dictionary
        :param pool_dict:
        :return:
        """
        assert pool_dict is not None
        pool_dict.pop(ABCPropertyGraph.FIELD_DELEGATION, None)
        self.pool_details = pool_dict

    def get_pool_details(self) -> Dict:
        """
        Return pool label or capacity details as a dictionary.
        :return:
        """
        return self.pool_details

    def set_delegation_id(self, *, delegation_id: str) -> None:
        assert delegation_id is not None
        self.delegation_id = delegation_id

    def add_defined_for(self, *, node_ids) -> None:
        """
        add a single element or a list
        :param node_ids: single node id or list
        :return:
        """
        assert node_ids is not None
        if isinstance(node_ids, list):
            self.for_.extend(node_ids)
        elif isinstance(node_ids, str):
            self.for_.append(node_ids)

    def is_defined_on(self, *, node_id: str) -> bool:
        """
        Is this pool defined on this node?
        :param node_id:
        :return:
        """
        return node_id == self.on_

    def is_defined_for(self, *, node_id: str) -> bool:
        """
        Is this pool defined for this node?
        :param node_id:
        :return:
        """
        return node_id in self.for_

    def get_defined_on(self) -> str:
        return self.on_

    def get_defined_for(self) -> List[str]:
        return self.for_

    def get_pool_id(self) -> str:
        return self.pool_id

    def get_pool_type(self) -> DelegationType:
        return self.type

    def get_delegation_id(self) -> str:
        return self.delegation_id

    def validate_pool(self):
        """
        Validate a pool to make sure delegation id is present, on, for and details are defined
        :return:
        """
        if self.delegation_id is None:
            raise PoolException(f"Pool {self.pool_id} does not have a delegation ID")
        if self.get_defined_on() is None:
            raise PoolException(f"Pool {self.pool_id} is not defined on any node")
        if self.get_defined_for() is None or len(self.get_defined_for()) == 0:
            raise PoolException(f"Pool {self.pool_id} is not mentioned on any nodes")
        if self.get_defined_on() not in self.get_defined_for():
            raise PoolException(f"Pool {self.pool_id} is not mentioned on the node where it is defined")
        if self.get_pool_details() is None:
            raise PoolException(f"Pool {self.pool_id} does not have any resource details")

    def __repr__(self) -> str:
        return f"{self.type} pool {self.pool_id} delegated to {self.delegation_id}: " \
               f"{self.on_}=> {self.for_} with {self.pool_details} "


class Pools:
    """
    Map between node ids, pools and delegations
    """
    def __init__(self, atype: DelegationType):
        self.pool_by_id = {}
        self.pools_by_delegation = None
        self.pool_type = atype

    def get_type(self) -> DelegationType:
        return self.pool_type

    def get_pool_by_id(self, *, pool_id: str, strict: bool = False) -> Pool:
        """
        Return a pool for this id. If strict is false (default) and pool doesn't exist,
        create a new pool of type matching this pools type.
        :param pool_id:
        :param strict
        :return:
        """
        assert pool_id is not None
        if strict:
            return self.pool_by_id.get(pool_id, None)
        else:
            p = self.pool_by_id.get(pool_id, None)
            if p is None:
                p = Pool(atype=self.pool_type, pool_id=pool_id)
                self.add_pool(pool=p)
            return p

    def build_index_by_delegation_id(self) -> None:
        """
        Validate all pools then index them by delegation ids, raise PoolException if any of them
        don't have the delegation id defined. Validate that all fields are filled in.
        :return:
        """
        self.pools_by_delegation = {}
        for pool in self.pool_by_id.values():
            pool.validate_pool()
            if self.pools_by_delegation.get(pool.get_delegation_id(), None) is None:
                self.pools_by_delegation[pool.get_delegation_id()] = []
            self.pools_by_delegation[pool.get_delegation_id()].append(pool)

    def get_pools_by_delegation_id(self, *, delegation_id: str) -> List[Pool]:
        """
        return a list of Pool(s) for a delegation. raises PoolException if the index based on
        delegation ids have not been built yet
        :param delegation_id:
        :return:
        """
        assert delegation_id is not None
        if self.pools_by_delegation is None:
            raise PoolException("Pools are node indexed by delegation, please run build_index_by_delegation_id()")
        return self.pools_by_delegation.get(delegation_id, None)

    def add_pool(self, *, pool: Pool) -> None:
        """
        add a pool to the index
        :param pool:
        :return:
        """
        assert pool is not None
        assert pool.get_pool_id() is not None
        self.pool_by_id[pool.get_pool_id()] = pool

    def get_delegation_ids(self) -> Set:
        """
        return a set of all delegation ids
        :return:
        """
        if self.pools_by_delegation is None:
            raise PoolException("Pools are not indexed by delegation, unable to get set of delegation ids")
        return set(self.pools_by_delegation.keys())

    def get_node_ids(self, *, delegation_id: str) -> Set:
        """
        Get a set of nodes for a given delegation across all pools
        :param delegation_id:
        :return:
        """
        assert delegation_id is not None
        if self.pools_by_delegation is None:
            raise PoolException("Pools are not indexed by delegation, unable to get node ids by delegation id")
        ret = set()
        if delegation_id not in self.pools_by_delegation.keys():
            return ret
        for pool in self.pools_by_delegation[delegation_id]:
            ret.update(pool.get_defined_for())
        return ret

    def validate_pools(self):
        """
        Make sure all pools have a defined on and for nodes.
        :return:
        """
        for _, pool in self.pool_by_id.items():
            pool.validate_pool()

    def __repr__(self):
        return f"{self.pool_by_id}"


class PoolException(Exception):
    """
    Exception with a pool
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Pool exception: {msg}")


class DelegationException(Exception):
    """
    Exception with a pool
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Delegation exception: {msg}")


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
from typing import List, Dict
import uuid
import networkx_query as nxq

from .abc_asm import ABCASMMixin
from ..abc_property_graph import ABCPropertyGraph, PropertyGraphQueryException
from ..networkx_property_graph import NetworkXPropertyGraph, NetworkXGraphImporter
from ...slivers.attached_components import ComponentSliver
from ...slivers.interface_info import InterfaceSliver, InterfaceInfo
from ...slivers.network_link import NetworkLinkSliver
from ...slivers.switch_fabric import SwitchFabricSliver
from ...slivers.network_node import NodeSliver as NodeSliver


class NetworkxASM(NetworkXPropertyGraph, ABCASMMixin):
    """
    Class implementing Abstract Slice Model on top of NetworkX
    """
    def __init__(self, *, graph_id: str, importer: NetworkXGraphImporter, logger=None):
        super().__init__(graph_id=graph_id, importer=importer, logger=logger)

    def get_all_network_nodes(self) -> List[str]:
        """
        Return a list of node ids of all graphs of type NetworkNode
        :return:
        """
        my_graph = self.storage.get_graph(self.graph_id)
        graph_nodes = list(nxq.search_nodes(my_graph,
                                            {'and': [
                                                {'eq': [ABCPropertyGraph.GRAPH_ID, self.graph_id]},
                                                {'eq': [ABCPropertyGraph.PROP_CLASS,
                                                        ABCPropertyGraph.CLASS_NetworkNode]}
                                            ]
                                            }))
        ret = list()
        for n in graph_nodes:
            ret.append(my_graph.nodes[n][ABCPropertyGraph.NODE_ID])
        return ret

    def get_all_network_links(self) -> List[str]:
        my_graph = self.storage.get_graph(self.graph_id)
        graph_links = list(nxq.search_nodes(my_graph,
                                            {'and': [
                                                {'eq': [ABCPropertyGraph.GRAPH_ID, self.graph_id]},
                                                {'eq': [ABCPropertyGraph.PROP_CLASS,
                                                        ABCPropertyGraph.CLASS_Link]}
                                            ]
                                            }))
        ret = list()
        for n in graph_links:
            ret.append(my_graph.nodes[n][ABCPropertyGraph.NODE_ID])
        return ret

    def get_all_link_interfaces(self, link_id: str) -> List[str]:
        assert link_id is not None
        # check this is a link
        labels, parent_props = self.get_node_properties(node_id=link_id)
        if ABCPropertyGraph.CLASS_Link not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=link_id,
                                              msg="Node type is not Link")
        return self.get_first_neighbor(node_id=link_id, rel=ABCPropertyGraph.REL_CONNECTS,
                                       node_label=ABCPropertyGraph.CLASS_ConnectionPoint)

    def get_all_network_node_components(self, parent_node_id: str) -> List[str]:
        """
        Return a list of components, children of a prent (presumably network node)
        :param parent_node_id:
        :return:
        """
        assert parent_node_id is not None
        # check that parent is a NetworkNode
        labels, parent_props = self.get_node_properties(node_id=parent_node_id)
        if ABCPropertyGraph.CLASS_NetworkNode not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=parent_node_id,
                                              msg="Parent node type is not NetworkNode")
        return self.get_first_neighbor(node_id=parent_node_id, rel=ABCPropertyGraph.REL_HAS,
                                       node_label=ABCPropertyGraph.CLASS_Component)

    def get_all_network_node_or_component_sfs(self, parent_node_id: str) -> List[str]:
        assert parent_node_id is not None
        # check that parent is a NetworkNode or Component
        labels, parent_props = self.get_node_properties(node_id=parent_node_id)
        if ABCPropertyGraph.CLASS_NetworkNode not in labels and \
            ABCPropertyGraph.CLASS_Component not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=parent_node_id,
                                              msg="Parent node type is not NetworkNode or Component")
        return self.get_first_neighbor(node_id=parent_node_id, rel=ABCPropertyGraph.REL_HAS,
                                       node_label=ABCPropertyGraph.CLASS_SwitchFabric)

    def get_all_node_or_component_connection_points(self, parent_node_id: str) -> List[str]:
        """
        Get a list of interfaces attached via switch fabrics
        :param parent_node_id:
        :return:
        """
        assert parent_node_id is not None
        labels, parent_props = self.get_node_properties(node_id=parent_node_id)
        if ABCPropertyGraph.CLASS_NetworkNode not in labels and \
                ABCPropertyGraph.CLASS_Component not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=parent_node_id,
                                              msg="Parent node type is not NetworkNode or Component")
        sfs_ifs = self.get_first_and_second_neighbor(node_id=parent_node_id, rel1=ABCPropertyGraph.REL_HAS,
                                                     node1_label=ABCPropertyGraph.CLASS_SwitchFabric,
                                                     rel2=ABCPropertyGraph.REL_CONNECTS,
                                                     node2_label=ABCPropertyGraph.CLASS_ConnectionPoint)
        ret = list()
        # return only interface IDs, not interested in SwitchFabrics
        for tup in sfs_ifs:
            ret.append(tup[1])
        return ret

    def get_all_sf_connection_points(self, parent_node_id: str) -> List[str]:
        assert parent_node_id is not None
        # check that parent is a SwitchFabric
        labels, parent_props = self.get_node_properties(node_id=parent_node_id)
        if ABCPropertyGraph.CLASS_SwitchFabric not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=parent_node_id,
                                              msg="Parent node type is not SwitchFabric")
        return self.get_first_neighbor(node_id=parent_node_id, rel=ABCPropertyGraph.REL_CONNECTS,
                                       node_label=ABCPropertyGraph.CLASS_ConnectionPoint)

    def find_node_by_name(self, node_name: str, label: str) -> str:
        """
        Return a node id of a node with this name
        :param node_name:
        :param label: label/class of the node
        :return:
        """
        assert node_name is not None
        my_graph = self.storage.get_graph(self.graph_id)
        graph_nodes = list(nxq.search_nodes(my_graph,
                                            {'and': [
                                                {'eq': [ABCPropertyGraph.GRAPH_ID, self.graph_id]},
                                                {'eq': [ABCPropertyGraph.PROP_NAME, node_name]},
                                                {'eq': [ABCPropertyGraph.PROP_CLASS, label]}
                                            ]}))
        if len(graph_nodes) == 0:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                              msg=f"Unable to find node with name {node_name} class {label}")
        if len(graph_nodes) > 1:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                              msg=f"Graph contains multiple nodes with name {node_name} class {label}")

        return my_graph.nodes[graph_nodes[0]][ABCPropertyGraph.NODE_ID]

    def check_node_unique(self, *, label: str, name: str):
        """
        Check no other node of this class/label and name exists
        :param label:
        :param name:
        :return:
        """
        graph_nodes = list(nxq.search_nodes(self.storage.get_graph(self.graph_id),
                                            {'and': [
                                                {'eq': [ABCPropertyGraph.GRAPH_ID, self.graph_id]},
                                                {'eq': [ABCPropertyGraph.PROP_NAME, name]},
                                                {'eq': [ABCPropertyGraph.PROP_CLASS, label]}
                                            ]}))
        return len(graph_nodes) == 0

    def find_component_by_name(self, *, parent_node_id: str, component_name: str) -> str:

        component_id_list = self.get_all_network_node_components(parent_node_id=parent_node_id)
        for cid in component_id_list:
            _, cprops = self.get_node_properties(node_id=cid)
            if cprops[ABCPropertyGraph.PROP_NAME] == component_name:
                return cid
        raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                          msg=f"Unable to find component with name {component_name}")

    def find_sf_by_name(self, *, parent_node_id: str, sfname: str) -> str:

        sf_id_list = self.get_all_network_node_or_component_sfs(parent_node_id=parent_node_id)
        for cid in sf_id_list:
            _, cprops = self.get_node_properties(node_id=cid)
            if cprops[ABCPropertyGraph.PROP_NAME] == sfname:
                return cid
        raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                          msg=f"Unable to find SwitchFabric with name {sfname}")

    def find_connection_point_by_name(self, *, parent_node_id: str, iname: str) -> str:

        if_id_list = self.get_all_sf_connection_points(parent_node_id=parent_node_id)
        for cid in if_id_list:
            _, cprops = self.get_node_properties(node_id=cid)
            if cprops[ABCPropertyGraph.PROP_NAME] == iname:
                return cid
        raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                          msg=f"Unable to find ConnectionPoint with name {iname}")

    def remove_network_node_with_components_sfs_cps_and_links(self, node_id: str):
        """
        Remove a network node, all of components and their interfaces, parent interfaces
        and connected links (as appropriate)
        :param node_id:
        :return:
        """
        # the network node itself
        # its components and switch fabrics
        # component switch fabrics
        # switch fabric interfaces (if present)
        # their parent interfaces (if present)
        # the 'Link' objects connected to interfaces if only one or no other interface connects to them

        # check we are a network node
        labels, node_props = self.get_node_properties(node_id=node_id)
        if ABCPropertyGraph.CLASS_NetworkNode not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_id,
                                              msg="This node type is not NetworkNode")
        # get a list of components
        components = self.get_first_neighbor(node_id=node_id, rel=ABCPropertyGraph.REL_HAS,
                                             node_label=ABCPropertyGraph.CLASS_Component)
        for cid in components:
            self.remove_component_with_sfs_cps_and_links(node_id=cid)

        # get a list of switch fabrics, delete the node
        switch_fabrics = self.get_first_neighbor(node_id=node_id, rel=ABCPropertyGraph.REL_HAS,
                                                 node_label=ABCPropertyGraph.CLASS_SwitchFabric)
        self.delete_node(node_id=node_id)
        for cid in switch_fabrics:
            self.remove_sf_with_cps_and_links(node_id=cid)

    def remove_component_with_sfs_cps_and_links(self, node_id: str):
        """
        Remove a component of a network node with all attached elements (switch fabrics
        and interfaces). Parent interfaces and links are removed if they have no other
        children or connected interfaces.
        :param node_id: component node id
        :return:
        """
        # check we are a component
        labels, node_props = self.get_node_properties(node_id=node_id)
        if ABCPropertyGraph.CLASS_Component not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_id,
                                              msg="This node type is not Component")
        # get a list of SwitchFabrics, delete the node
        switch_fabrics = self.get_first_neighbor(node_id=node_id, rel=ABCPropertyGraph.REL_HAS,
                                                 node_label=ABCPropertyGraph.CLASS_SwitchFabric)
        self.delete_node(node_id=node_id)
        for sf in switch_fabrics:
            self.remove_sf_with_cps_and_links(node_id=sf)

    def remove_network_link(self, node_id: str):

        # check we are a link
        labels, node_props = self.get_node_properties(node_id=node_id)
        if ABCPropertyGraph.CLASS_Link not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_id,
                                              msg="This node type is not Link")
        # edges to interfaces/connection points will disappear
        self.delete_node(node_id=node_id)

    def remove_sf_with_cps_and_links(self, node_id: str):
        """
        Remove switch fabric with subtending interfaces and links
        :param node_id:
        :return:
        """
        # check we are a switch fabric
        labels, node_props = self.get_node_properties(node_id=node_id)
        if ABCPropertyGraph.CLASS_SwitchFabric not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_id,
                                              msg="This node type is not SwitchFabric")
        # get a list of interfaces, delete the switch fabric
        interfaces = self.get_first_neighbor(node_id=node_id, rel=ABCPropertyGraph.REL_CONNECTS,
                                             node_label=ABCPropertyGraph.CLASS_ConnectionPoint)
        self.delete_node(node_id=node_id)
        for iif in interfaces:
            self.remove_cp_and_links(node_id=iif)

    def remove_cp_and_links(self, node_id: str):
        """
        Remove ConnectionPoint and links. Parent ConnectionPoints and links are removed
        if they have no other children or other connected connection points.
        :param node_id: interface node id
        :return:
        """
        # some interfaces may have parent interfaces, which can be deleted if they only have the
        # one child
        interfaces_to_delete = {node_id}
        parents = self.get_first_neighbor(node_id=node_id,
                                          rel=ABCPropertyGraph.REL_CONNECTS,
                                          node_label=ABCPropertyGraph.CLASS_ConnectionPoint)
        for parent in parents:  # really should only be one parent interface
            children = self.get_first_neighbor(node_id=parent,
                                               rel=ABCPropertyGraph.REL_CONNECTS,
                                               node_label=ABCPropertyGraph.CLASS_ConnectionPoint)
            if len(children) == 1:  # if only child, can delete parent
                interfaces_to_delete.add(parent)
        # interfaces themselves and parent interfaces
        # may be connected to links which can be deleted if nothing
        # else connects to them
        links_to_delete = set()
        for interface in interfaces_to_delete:
            links = self.get_first_neighbor(node_id=interface,
                                            rel=ABCPropertyGraph.REL_CONNECTS,
                                            node_label=ABCPropertyGraph.CLASS_Link)
            for link in links:  # should only be one
                connected_interfaces = self.get_first_neighbor(node_id=link,
                                                               rel=ABCPropertyGraph.REL_CONNECTS,
                                                               node_label=ABCPropertyGraph.CLASS_ConnectionPoint)
                if len(connected_interfaces) == 2:  # connected to us and another thing, can delete
                    links_to_delete.add(link)
        # delete nodes in these sets
        for deleted_id in interfaces_to_delete.union(links_to_delete):
            self.delete_node(node_id=deleted_id)

    def add_network_node_sliver(self, *, sliver: NodeSliver):

        assert sliver is not None
        assert sliver.node_id is not None
        assert self.check_node_unique(label=ABCPropertyGraph.CLASS_NetworkNode, name=sliver.resource_name)

        props = self.node_sliver_to_graph_properties_dict(sliver)
        self.add_node(node_id=sliver.node_id, label=ABCPropertyGraph.CLASS_NetworkNode, props=props)
        # if components aren't empty, add components, their switch fabrics and interfaces
        aci = sliver.attached_components_info
        if aci is not None:
            for csliver in aci.devices.values():
                self.add_component_sliver(parent_node_id=sliver.node_id,
                                          component=csliver)
        # if switch fabrics arent empty add them with their interfaces
        sfi = sliver.switch_fabric_info
        if sfi is not None:
            for sf in sfi.switch_fabrics.values():
                self.add_switch_fabric_sliver(parent_node_id=sliver.node_id,
                                              switch_fabric=sf)

    def add_network_link_sliver(self, *, lsliver: NetworkLinkSliver, interfaces: List[str]):

        assert lsliver is not None
        assert lsliver.node_id is not None
        assert self.check_node_unique(label=ABCPropertyGraph.CLASS_Link, name=lsliver.resource_name)
        assert interfaces is not None

        props = self.link_sliver_to_graph_properties_dict(lsliver)
        self.add_node(node_id=lsliver.node_id, label=ABCPropertyGraph.CLASS_Link, props=props)
        # add edge links to specified interfaces
        for i in interfaces:
            self.add_link(node_a=lsliver.node_id, rel=ABCPropertyGraph.REL_CONNECTS, node_b=i)

    def add_component_sliver(self, *, parent_node_id: str, component: ComponentSliver):
        """
        Add network node component as a sliver, linking back to parent
        :param parent_node_id:
        :param component:
        :return:
        """
        assert component.node_id is not None
        assert parent_node_id is not None
        assert self.check_node_unique(label=ABCPropertyGraph.CLASS_Component, name=component.resource_name)
        props = self.component_sliver_to_graph_properties_dict(component)
        self.add_node(node_id=component.node_id, label=ABCPropertyGraph.CLASS_Component, props=props)
        self.add_link(node_a=parent_node_id, rel=ABCPropertyGraph.REL_HAS, node_b=component.node_id)
        sfi = component.switch_fabric_info
        if sfi is not None:
            for sf in sfi.switch_fabrics.values():
                self.add_switch_fabric_sliver(parent_node_id=component.node_id,
                                              switch_fabric=sf)

    def add_switch_fabric_sliver(self, *, parent_node_id: str, switch_fabric: SwitchFabricSliver):
        """
        Add switch fabric as a sliver to either node or component, linking back to parent.
        :param parent_node_id:
        :param switch_fabric:
        :return:
        """
        assert parent_node_id is not None
        assert switch_fabric.node_id is not None
        assert self.check_node_unique(label=ABCPropertyGraph.CLASS_SwitchFabric,
                                      name=switch_fabric.resource_name)
        props = self.switch_fabric_sliver_to_graph_properties_dict(switch_fabric)
        self.add_node(node_id=switch_fabric.node_id, label=ABCPropertyGraph.CLASS_SwitchFabric, props=props)
        self.add_link(node_a=parent_node_id, rel=ABCPropertyGraph.REL_HAS, node_b=switch_fabric.node_id)
        ii = switch_fabric.interface_info
        if ii is not None:
            for i in ii.interfaces.values():
                self.add_interface_sliver(parent_node_id=switch_fabric.node_id,
                                          interface=i)

    def add_interface_sliver(self, *, parent_node_id: str, interface: InterfaceSliver):
        """
        Add interface to a switch fabric, linking back to parent.
        :param parent_node_id: switch fabric id
        :param interface: interface sliver description
        :return:
        """
        assert interface.node_id is not None
        assert parent_node_id is not None
        assert self.check_node_unique(label=ABCPropertyGraph.CLASS_ConnectionPoint, name=interface.resource_name)
        props = self.interface_sliver_to_graph_properties_dict(interface)
        self.add_node(node_id=interface.node_id, label=ABCPropertyGraph.CLASS_ConnectionPoint, props=props)
        self.add_link(node_a=parent_node_id, rel=ABCPropertyGraph.REL_CONNECTS, node_b=interface.node_id)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)

from typing import Dict, Any, List

import uuid

from .model_element import ModelElement, ElementType
from .component import Component, ComponentType
from .switch_fabric import SwitchFabric
from .interface import Interface

from ..graph.abc_property_graph import ABCPropertyGraph

from ..slivers.network_node import NodeSliver
from ..slivers.network_node import NodeType
from ..slivers.switch_fabric import SFLayer


class Node(ModelElement):
    """
    A basic node of the topology
    """

    def __init__(self, *, name: str, node_id: str = None, topo: Any, etype: ElementType = ElementType.EXISTING,
                 ntype: NodeType = None, site: str = None, **kwargs):
        """
        Don't call this method yourself, call topology.add_node()
        node_id will be generated if not provided for experiment topologies

        :param name:
        :param node_id:
        :param topo:
        :param etype: is this supposed to exist or new should be created
        :param ntype: node type if it is new
        :param site: node site
        :param kwargs: any additional properties
        """
        assert name is not None
        assert topo is not None

        if etype == ElementType.NEW:
            # cant use isinstance as it would create circular import dependencies
            # node id myst be specified for new nodes in substrate topologies
            if str(topo.__class__) == "<class 'fim.user.topology.SubstrateTopology'>" and \
                    node_id is None:
                raise RuntimeError("When adding new nodes to substrate topology nodes you must specify static Node ID")
            if node_id is None:
                node_id = str(uuid.uuid4())
            super().__init__(name=name, node_id=node_id, topo=topo)
            if ntype is None:
                raise RuntimeError("When creating nodes you must specify NodeType")
            if site is None:
                raise RuntimeError("When creating nodes you must specify site")

            self.node_id = node_id
            sliver = NodeSliver()
            sliver.node_id = self.node_id
            sliver.set_resource_name(self.name)
            sliver.set_resource_type(ntype)
            sliver.set_site(site)
            sliver.set_properties(**kwargs)

            self.topo.graph_model.add_network_node_sliver(sliver=sliver)
        else:
            super().__init__(name=name, node_id=node_id, topo=topo)
            # check that this node exists
            existing_node_id = self.topo.graph_model.find_node_by_name(node_name=name,
                                                                       label=str(ABCPropertyGraph.CLASS_NetworkNode))
            if node_id is not None and existing_node_id != node_id:
                raise RuntimeError("Existing node id does not match provided. "
                                   "In general you shouldn't need to specify node id for existing nodes.")
            self.node_id = node_id

    def get_property(self, pname: str) -> Any:
        """
        Retrieve a node property
        :param pname:
        :return:
        """
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        node_sliver = self.topo.graph_model.node_sliver_from_graph_properties_dict(node_properties)
        return node_sliver.get_property(pname)

    def set_property(self, pname: str, pval: Any):
        """
        Set a node property
        :param pname:
        :param pval:
        :return:
        """
        node_sliver = NodeSliver()
        node_sliver.set_property(prop_name=pname, prop_val=pval)
        # write into the graph
        prop_dict = self.topo.graph_model.node_sliver_to_graph_properties_dict(node_sliver)
        self.topo.graph_model.update_node_properties(node_id=self.node_id, props=prop_dict)

    def set_properties(self, **kwargs):
        """
        Set multiple properties of the node
        :param kwargs:
        :return:
        """
        node_sliver = NodeSliver()
        node_sliver.set_properties(**kwargs)
        # write into the graph
        prop_dict = self.topo.graph_model.node_sliver_to_graph_properties_dict(node_sliver)
        self.topo.graph_model.update_node_properties(node_id=self.node_id, props=prop_dict)

    @staticmethod
    def list_properties() -> List[str]:
        return NodeSliver.list_properties()

    def add_component(self, *, name: str, node_id: str = None, ctype: ComponentType,
                      model: str,  switch_fabric_node_id: str = None, **kwargs) -> Component:
        """
        Add a component of specified type, model and name to this node. When working with substrate
        topologies you must specify the switch_fabric_node_id and provide a list of interface node ids.
        :param name:
        :param node_id:
        :param ctype:
        :param model:
        :param name:
        :param switch_fabric_node_id:
        :param kwargs: additional properties of the component
        :return:
        """
        # add component node and populate properties
        c = Component(name=name, node_id=node_id, topo=self.topo, etype=ElementType.NEW,
                      ctype=ctype, model=model, switch_fabric_node_id=switch_fabric_node_id,
                      parent_node_id=self.node_id, **kwargs)
        return c

    def add_switch_fabric(self, *, name: str, node_id: str = None, layer: SFLayer):
        """
        Add a switch fabric to node (mostly needed in substrate topologies)
        :param name:
        :param node_id:
        :param layer:
        :return:
        """
        sf = SwitchFabric(name=name, node_id=node_id, layer=layer, parent_node_id=self.node_id,
                          etype=ElementType.NEW, topo=self.topo)
        return sf

    def remove_component(self, name: str) -> None:
        """
        Remove a component from the node (and switch fabrics and their interfaces)
        :param name:
        :return:
        """
        assert name is not None
        node_id = self.topo.graph_model.find_component_by_name(parent_node_id=self.node_id,
                                                               component_name=name)
        self.topo.graph_model.remove_component_with_sfs_cps_and_links(node_id=node_id)

    def remove_switch_fabric(self, name: str) -> None:
        """
        Remove a switch fabric from the node (and all its interfaces)
        :param name:
        :return:
        """
        assert name is not None

        node_id = self.topo.graph_model.find_sf_by_name(parent_node_id=self.node_id,
                                                        sfname=name)
        self.topo.graph_model.remove_sf_with_cps_and_links(node_id=node_id)

    def __get_component_by_name(self, name: str) -> Component:
        """
        Find component of a node by its name, return Component object
        :param name:
        :return:
        """
        assert name is not None
        node_id = self.topo.graph_model.find_component_by_name(parent_node_id=self.node_id, component_name=name)
        return Component(name=name, node_id=node_id, topo=self.topo)

    def __get_component_by_id(self, node_id: str) -> Component:
        """
        Get component of a node by its node_id, return Component object
        :param node_id:
        :return:
        """
        assert node_id is not None
        _, node_props = self.topo.graph_model.get_node_properties(node_id=node_id)
        assert node_props.get(ABCPropertyGraph.PROP_NAME, None) is not None
        return Component(name=node_props[ABCPropertyGraph.PROP_NAME], node_id=node_id,
                         topo=self.topo)

    def __get_sf_by_id(self, node_id: str) -> SwitchFabric:
        """
        Get a switch fabric of a node by its node_id, return SwitchFabric object
        :param node_id:
        :return:
        """
        assert node_id is not None
        _, node_props = self.topo.graph_model.get_node_properties(node_id=node_id)
        assert node_props.get(ABCPropertyGraph.PROP_NAME, None) is not None
        return SwitchFabric(name=node_props[ABCPropertyGraph.PROP_NAME], node_id=node_id,
                            topo=self.topo)

    def __get_interface_by_id(self, node_id: str) -> Interface:
        """
        Get an interface of a node by its node_id, return Interface object
        :param node_id:
        :return:
        """
        assert node_id is not None
        _, node_props = self.topo.graph_model.get_node_properties(node_id=node_id)
        assert node_props.get(ABCPropertyGraph.PROP_NAME, None) is not None
        return Interface(name=node_props[ABCPropertyGraph.PROP_NAME], node_id=node_id,
                         topo=self.topo)

    def __list_components(self) -> Dict[str, Component]:
        """
        List all Components children of a node in the topology as a dictionary
        organized by component name. Modifying the dictionary will not affect
        the underlying model, but modifying Components in the dictionary will.
        :return:
        """
        node_id_list = self.topo.graph_model.get_all_network_node_components(parent_node_id=self.node_id)
        # Could consider using frozendict or other immutable idioms
        ret = dict()
        for nid in node_id_list:
            c = self.__get_component_by_id(nid)
            ret[c.name] = c
        return ret

    def __list_switch_fabrics(self) -> Dict[str, SwitchFabric]:
        """
        List all switch fabric children of a node as a dictionary organized
        by switch fabric name. Modifying the dictionary will not affect
        the underlying model, but modifying Components in the dictionary will.
        :return:
        """
        node_id_list = self.topo.graph_model.get_all_network_node_or_component_sfs(parent_node_id=self.node_id)
        # Could consider using frozendict or other immutable idioms
        ret = dict()
        for nid in node_id_list:
            c = self.__get_sf_by_id(nid)
            ret[c.name] = c
        return ret

    def __list_direct_interfaces(self) -> Dict[str, Interface]:
        """
        List all directly-attached interfaces of the node as a dictionary
        :return:
        """
        # immediately-attached interfaces
        node_if_list = self.topo.graph_model.get_all_node_or_component_connection_points(parent_node_id=self.node_id)
        # Could consider using frozendict here
        ret = dict()
        for nid in node_if_list:
            i = self.__get_interface_by_id(nid)
            ret[i.name] = i
        return ret

    def __list_interfaces(self) -> Dict[str, Interface]:
        """
        List all interfaces of node and its components
        :return:
        """
        direct_interfaces = self.__list_direct_interfaces()
        cdict = self.__list_components()
        for k, v in cdict.items():
            comp_interfaces = v.interfaces
            direct_interfaces.update(comp_interfaces)
        return direct_interfaces

    def get_component(self, name: str):
        """
        Get a component by this name
        :param name:
        :return:
        """
        return self.__get_component_by_name(name)

    def __getattr__(self, item):
        """
        Special handling for attributes like 'components' and 'interfaces' -
        which query into the model. They return dicts and list
        containers. Modifying containers does not affect the underlying
        graph mode, but modifying elements of lists or values of dicts does.
        :param item:
        :return:
        """
        if item == 'components':
            return self.__list_components()
        if item == 'interfaces':
            return self.__list_interfaces()
        if item == 'direct_interfaces':
            return self.__list_direct_interfaces()
        if item == 'switch_fabrics':
            return self.__list_switch_fabrics()

    def __repr__(self):
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        node_sliver = self.topo.graph_model.node_sliver_from_graph_properties_dict(node_properties)
        return node_sliver.__repr__()

    def __str__(self):
        return self.__repr__()





#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
A pluggable interface for externally supplied implementations
for various algorithms operating on FIM. Examples include:
- Responding to controller queries and producing a BQM using CBM
- Partitioning slice model into slivers
- Implementing various modify operations
Various classes in FIM have hooks to call registered external implementations
at well-defined points. Any registered implementation subsumes the native FIM
implementation of the corresponding functionality. In case of multiple registered
implementations the last one takes precedence.
"""
from abc import ABC, abstractmethod
import inspect
from typing import List
from enum import Enum

from .graph.abc_property_graph import ABCPropertyGraph


class PluggableType(Enum):
    AM = 1
    Broker = 2
    Orchestrator = 3


class ABCPluggable(ABC):
    """
    Defines framework that can be used to override default behaviors in different
    parts of FIM. Such methods begin with 'plug_'
    A pluggable class doesn't have to implement all available methods nor does it have
    to inherit from this class or its subclasses.
    A single instance of a class implementing pluggable methods is created within a
    particular context (AM, Broker or Orchestrator), allowing in-memory state
    to be passed. Persistent state is derived from the attached Neo4j instance, currently
    a relational database interface would have to be implemented explicitly -
    extensions to support relational persistent storage abstractions for Pluggables
    may be added at a later time.
    """
    @classmethod
    def get_pluggable_methods(cls, detail: bool=True) -> List:
        """
        Return a list of tuples about pluggable methods that can be implemented.
        Each tuple is (method_name, signature, docstring).
        If detail is set to False, a simple list of names is returned
        :param detail:
        :return:
        """
        ret = list()
        for i in inspect.getmembers(cls):
            # returns tuples
            # ('__abstractmethods__', frozenset({'plug_partition_slice', 'plug_produce_bqm'}))
            # ('plug_produce_bqm', < function ABCPluggable.plug_produce_bqm at 0x105c8cd40 >)
            if i[0].startswith('plug_'):
                method = i[0]
                _callable = getattr(cls, method)
                docstring = inspect.getdoc(_callable)
                signature = inspect.signature(_callable)
                if detail:
                    ret.append((method, str(signature), docstring))
                else:
                    ret.append(method)
        return ret

    @staticmethod
    def get_implemented_methods(cls) -> List:
        """
        Get the pluggable methods a class implements as a list of method names.
        :param cls:
        :return:
        """
        ret = list()
        for i in inspect.getmembers(cls):
            # returns tuples
            # ('__abstractmethods__', frozenset({'plug_partition_slice', 'plug_produce_bqm'}))
            # ('plug_produce_bqm', < function ABCPluggable.plug_produce_bqm at 0x105c8cd40 >)
            if i[0].startswith('plug_'):
                method = i[0]
                ret.append(method)
        return ret


class BrokerPluggable(ABCPluggable):
    """
        Overridable Broker-specific methods for graph/resource manipulations in FIM
    """
    PLUGGABLE_PRODUCE_BQM = 'plug_produce_bqm'

    @abstractmethod
    def plug_produce_bqm(self, *, cbm: ABCPropertyGraph, **kwargs) -> ABCPropertyGraph:
        """
        Produce a Broker Query Model based on a Combined Broker Model and
        some unspecified query parameters
        :param cbm:
        :param kwargs:
        :return:
        """


class AMPluggable(ABCPluggable):
    """
    Overridable AM-specific methods for graph/resource manipulations in FIM
    """


class OrchestratorPluggable(ABCPluggable):
    """
    Overridable Controller-specific methods for graph/resource manipulations in FIM
    """
    PLUGGABLE_PARTITION_SLICE = 'plug_partition_slice'

    @abstractmethod
    def plug_partition_slice(self, *, slice: ABCPropertyGraph, **kwargs) -> List:
        """
        Partition slice into slivers using a slice model and some
        unspecified other parameters
        :param slice:
        :param kwargs:
        :return:
        """


class PluggableRegistry:
    """
    Singleton which accepts and stores a registration for external
    implementation defined by Pluggable. Those wishing to override
    default FIM behaviors should see the return of Pluggable.get_pluggable_methods(),
    implement desired methods in their class and register it with PluggableRegistry.
    An instance of this class will be created as a singleton and the methods
    will be invoked at appropriate points within FIM.
    A pluggable class does not have to inherit from Pluggable.
    """
    PLUGGABLE_MAP = {PluggableType.AM: (AMPluggable,
                                        set(AMPluggable.get_pluggable_methods(detail=False))),
                     PluggableType.Orchestrator: (OrchestratorPluggable,
                                                  set(OrchestratorPluggable.get_pluggable_methods(detail=False))),
                     PluggableType.Broker: (BrokerPluggable,
                                            set(BrokerPluggable.get_pluggable_methods(detail=False)))}

    class __Pluggable:
        def __init__(self, t: PluggableType, p):
            self.pluggable = p
            # figure out which methods it implements
            pluggable_type, pluggable_methods = PluggableRegistry.PLUGGABLE_MAP[t]
            self.implemented_methods = pluggable_type.get_implemented_methods(p)
            if len(self.implemented_methods) == 0 or \
                    len(pluggable_methods.intersection(self.implemented_methods)) == 0:
                raise RuntimeError(f"Plugin {p} does not implement any of the expected methods: "
                                   f"{pluggable_methods}")
            self.pluggable_instance = p()

        def get_implemented_methods(self):
            return self.implemented_methods

        def get_method_callable(self, method):
            if method not in self.implemented_methods:
                raise RuntimeError(f"Method {method} is not implemented by the plugin {self.pluggable}")
            return getattr(self.pluggable_instance, method)

    instance = dict()

    def __init__(self):
        pass

    def register_pluggable(self, *, t: PluggableType, p):
        """
        Register a pluggable interface
        :param t:
        :param p:
        :return:
        """
        if PluggableRegistry.instance.get(t, None) is None:
            PluggableRegistry.instance[t] = PluggableRegistry.__Pluggable(t, p)
        else:
            raise RuntimeError(f"Another plugin {PluggableRegistry.instance[t].pluggable} "
                               f"class is already registered for type {t.name}")

    def pluggable_registered(self, *, t: PluggableType) -> bool:
        """
        Return true if a pluggable already is registered
        :param t:
        :return:
        """
        return PluggableRegistry.instance.get(t, None) is not None

    def get_implemented_methods(self, *, t: PluggableType) -> List:
        """
        Get implemented methods for a pluggable type as a list of strings
        :param t:
        :return:
        """
        return PluggableRegistry.instance[t].get_implemented_methods()

    def get_method_callable(self, *, t: PluggableType, method):
        """
        Get a callable for a given method name
        :param t:
        :param method:
        :return:
        """
        return PluggableRegistry.instance[t].get_method_callable(method=method)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
from typing import List
import json

from abc import ABC, abstractmethod


class JSONField(ABC):

    @abstractmethod
    def set_fields(self, **kwargs):
        """
        Set the fields of this object
        :param kwargs:
        :return:
        """

    def to_json(self) -> str or None:
        """
        Dumps to JSON the __dict__ of the instance. Be careful as the fields in this
        class should only be those that can be present in JSON output.
        If there are no values in the object, returns None
        :return:
        """
        d = self.__dict__.copy()
        for k in self.__dict__:
            if d[k] is None or d[k] == 0:
                d.pop(k)
        return json.dumps(d, skipkeys=True, sort_keys=True)

    def from_json(self, json_string: str):
        """
        Set fields from json string (if string is none or empty,
        nothing happens)
        :param json_string:
        :return: self for call chaining
        """
        if json_string is None or len(json_string) == 0:
            return self
        d = json.loads(json_string)
        self.set_fields(**d)
        return self

    def __repr__(self):
        return self.to_json()

    def __str__(self):
        return self.to_json()

    def list_fields(self):
        l = list(self.__dict__.keys())
        l.sort()
        return l


class Capacities(JSONField):
    """
    Implements basic capacity field handling - encoding and decoding
    from JSON dictionaries of properties
    """
    def __init__(self):
        self.cpu = 0
        self.core = 0
        self.ram = 0
        self.disk = 0
        self.bw = 0
        self.unit = 0

    def set_fields(self, **kwargs):
        """
        Universal integer setter for all fields.
        Values should be non-negative integers. Throws a RuntimeError
        if you try to set a non-existent field.
        :param kwargs:
        :return: self to support call chaining
        """
        for k, v in kwargs.items():
            assert v >= 0
            try:
                # will toss an exception if field is not defined
                self.__getattribute__(k)
                self.__setattr__(k, v)
            except AttributeError:
                raise RuntimeError(f"Unable to set field {k} of capacity, no such field available")
        return self


class Labels(JSONField):
    """
    Class implementing various encodings of labels field, encoding
    and decoding from JSON dictionaries of properties
    """
    def __init__(self):
        self.bdf = None
        self.mac = None
        self.ipv4 = None
        self.ipv4_range = None
        self.ipv6 = None
        self.ipv6_range = None
        self.asn = None
        self.vlan = None
        self.vlan_range = None
        self.node = None

    def set_fields(self, **kwargs):
        """
        Universal setter for all fields (just strip the l_ from the field
        name of the field). Values should be strings or lists of strings.
        Throws a RuntimeError if you try to set a non-existent field.
        :param kwargs:
        :return: self to support call chaining
        """
        for k, v in kwargs.items():
            assert v is not None  # could be strings or lists of strings
            assert isinstance(v, str) or isinstance(v, list)
            try:
                # will toss an exception if field is not defined
                self.__getattribute__(k)
                self.__setattr__(k, v)
            except AttributeError:
                raise RuntimeError(f"Unable to set field {k} of labels, no such field available")
        # to support call chaining
        return self



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import ipaddress
import uuid
import enum

from .base_sliver import BaseSliver


class InterfaceType(enum.Enum):
    AccessPort = enum.auto()
    TrunkPort = enum.auto()
    vInt = enum.auto()

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name


class InterfaceSliver(BaseSliver):

    def __init__(self):
        # addresses are labels, no need for additional fields
        super().__init__()

    @staticmethod
    def type_from_str(ntype: str) -> InterfaceType:
        if ntype is None:
            return None
        for t in InterfaceType:
            if ntype == str(t):
                return t


class InterfaceInfo:
    def __init__(self):
        self.interfaces = {}

    def add_interface(self, interface_info: InterfaceSliver):
        self.interfaces[interface_info.resource_name] = interface_info

    def remove_interface(self, name: str):
        if name in self.interfaces.keys():
            self.interfaces.pop(name)

    def get_interface(self, name: str):
        return self.interfaces.get(name, None)
